class AwsaccountsController < ApplicationController
  require 'utility'

  before_action :set_remote_user, :set_page_title, :set_breadcrumbs
  before_action :set_paper_trail_whodunnit
  before_action :set_roles
  before_action :allow_access
  before_action :set_awsaccount, only: [:show, :edit, :update, :destroy]

  # GET /awsaccounts
  def index
    # Display the "List ALL Configurations" only when user is some kind of
    # admin.
    @show_all_button = (@is_admin or @is_readonly_admin)
  end

  def list
    username = Utility.get_REMOTE_USER()

    # Only show objects owned by REMOTE_USER or objects with a group
    # that REMOTE_USER is in. However, we allow
    # admins to see all the entries if @is_admin is true.
    if (params[:search]) then
      search_string = params[:search].to_s()
      @awsaccounts = Awsaccount.where("number LIKE ?",
                                  "%#{search_string}%")

      # Remove any that this user should not have access to.
      @awsaccounts = Awsaccount.return_only_accessible(@awsaccounts, username)

      @breadcrumbs.push(
        ["All AWS Account(s) (search string &quot;#{search_string}&quot;)".html_safe, ""]
      )
    else
      user       = Utility.get_REMOTE_USER()
      @awsaccounts = Awsaccount.this_users_awsaccounts(user)
      @breadcrumbs.push(
        ["AWS Account(s) for #{username}", ""]
      )
    end
    flash.notice = "Search found #{@awsaccounts.size} AWS Account(s)"
  end

  # GET /awsaccounts/all
  # Only for admins.
  def all
    username = Utility.get_REMOTE_USER()

    if (@is_admin or @is_readonly_admin) then
      @awsaccounts = Awsaccount.all()
      @breadcrumbs.push(
        ["All SP Configuration(s) (admin)", ""]
      )
      flash.notice = "#{@awsaccounts.size} AWS Account(s)"
      render 'list'
    else
      redirect_to "/awsaccounts/list"
    end
  end


  # GET /awsaccounts/1
  # GET /awsaccounts/1.json
  def show
    @breadcrumbs.push(
      ["Account #{@awsaccount.number}", awsaccount_path(@awsaccount.id)]
    )

    @group_string = @awsaccount.workgroup_names.join(', ')

    if (@group_string == '') then
      @group_string = "no owning workgroups"
    end

    @role_mappings_string = ''
    @awsaccount.awsrolemappings.each do |awsrolemapping|
      wg   = awsrolemapping.workgroup
      role = awsrolemapping.role
      @role_mappings_string << %Q[<tr><td class="awsrolemapping">#{wg}</td><td class="awsrolemapping">#{role}</td></tr>\n]
    end
  end

  # GET /awsaccounts/new
  def new
    @awsaccount = Awsaccount.new
    @breadcrumbs.push(
      ["New AWS Account", ""]
    )
  end

  # GET /awsaccounts/1/edit
  def edit
    @breadcrumbs.push(
      ["Edit Account #{@awsaccount.number}", ""]
    )
  end

  # POST /awsaccounts
  def create
    @awsaccount = Awsaccount.new(awsaccount_params)

    # Force the instance to be AWS (not GovCloud). We may change this
    # later if more people start using GovCloud.
    @awsaccount.instance = 'aws'

    # For reasons beyond my understanding, the owning groups attributes
    # from the submitted web form are not available in the model for
    # validation. But we need them to do proper validation, so to get
    # around this we pass them into the model using
    # 'add_groups_attributes', a method we wrote for this purpose.
    @awsaccount.add_groups_attributes(awsaccount_params['awsowninggroups_attributes'])
    Utility.logger.info("awsaccount_params is #{awsaccount_params.inspect}")

    respond_to do |format|
      if @awsaccount.save
        format.html { redirect_to @awsaccount, notice: 'successfully created',
                      alert: '(account mapping should be available within 15 minutes)' }
        format.json { render :show, status: :created, location: @awsaccount }
      else
        format.html { render :new }
        format.json { render json: @awsaccount.errors, status: :unprocessable_entity }
      end
    end

  end

  # PATCH/PUT /awsaccounts/1
  # PATCH/PUT /awsaccounts/1.json
  def update
    @awsaccount.attributes = awsaccount_params

    # For reasons beyond my understanding, the owning groups attributes
    # from the submitted web form are not available in the model for
    # validation. But we need them to do proper validation, so to get
    # around this we pass them into the model using
    # 'add_groups_attributes', a method we wrote for this purpose.
    @awsaccount.add_groups_attributes(awsaccount_params['awsowninggroups_attributes'])
    Utility.logger.info("awsaccount_params is #{awsaccount_params.inspect}")

    respond_to do |format|
      if @awsaccount.save()
        format.html { redirect_to @awsaccount, notice: 'successfully updated',
                      alert: '(changes should take effect within 15 minutes)' }
        format.json { render :show, status: :ok, location: @awsaccount }
      else
        format.html { render :edit }
        format.json { render json: @awsaccount.errors, status: :unprocessable_entity }
      end
    end
  end

  # DELETE /awsaccount/1
  # DELETE /awsaccount/1.json
  def destroy
    @awsaccount.destroy
    respond_to do |format|
      format.html { redirect_to awsaccounts_url, notice: 'AWS Account was successfully removed.' }
      format.json { head :no_content }
    end
  end

  private
    def set_page_title
      @page_title      = "AWS Accounts Mappings"
      @page_title_path = awsaccounts_path
    end

    # Use callbacks to share common setup or constraints between actions.
    def set_awsaccount
      @awsaccount = Awsaccount.find(params[:id])
      if (!can_access()) then
        redirect_to "/errors/forbidden_record"
        return
      end
    end

    def awsaccount_params
      params.require(:awsaccount).permit(:instance,
                                         :number,
                                         :master,
                                         :nickname,
                                         :notes,
                                         awsowninggroups_attributes: [:id, :workgroup, :_destroy],
                                        )
    end

    def set_breadcrumbs
      @breadcrumbs = Utility.first_level_breadcrumb
      @breadcrumbs.push(['My AWS Accounts', awsaccounts_path])
      @model_name = 'awsaccount'
    end

    def allow_access
      # We allow access if this application is "visible" to non-admins OR
      # the current user is an admin.
      if (not($visible_applications.include?('awsaccounts') or @is_admin)) then
        redirect_to '/403'
      end
    end

    # Does the current user (i.e., SUNetID) "own" this configuration?
    # That is, does the current user satisfy one (or more) of the
    # following:
    #
    #   1. user is an ADMIN
    #   2. user is in one of the owning groups
    #
    # If yes, return true, otherwise return false.
    def can_access()
      user = Utility.get_REMOTE_USER()

      # 1. user is an SPDB admin.
      if (Utility.admin?(user)) then
        return true
      end

      # 2. user is one (or more) of this models groups.
      if (@awsaccount.user_in_workgroups(user)) then
        return true
      end

      return false
    end


end
