class SpconfigsController < ApplicationController
  require 'utility'
  require 'open-uri'
  require 'nokogiri'
  require 'openssl'

  # Note: Some of these actions are defined in application_controller.rb
  before_action :set_remote_user, :get_total_number_of_configs
  before_action :set_paper_trail_whodunnit
  before_action :set_roles, :set_page_title, :set_breadcrumbs_spconfigs
  before_action :set_spconfig
#  before_action :set_default_federation
  after_action  :update_partners_list, only: [:create, :update, :destroy]

  # Skip the set_spconfig ...
  skip_before_action :set_spconfig, except: [:show, :edit, :update, :destroy, :attributes]

  # GET /spconfigs
  # GET /spconfigs.json
  # Show only a few buttons: create, search, show-all.
  def index
    # Display the "List ALL Configurations" only when user is some kind of
    # admin.
    @show_all_button = (@is_admin or @is_readonly_admin)

    # Display the "Add new SP configuration" only when user has write
    # privileges.
    @show_add_new_button = @is_writer
  end

  # GET /spconfigs/list
  # Show everything that the user can see. If there is search parameter, use
  # that to limit the results.
  def list
    username = Utility.get_REMOTE_USER()

    # Only show objects owned by REMOTE_USER or objects with a group
    # that REMOTE_USER is in. However, we allow
    # admins to see all the entries if @is_admin is true.
    if (params[:search]) then
      search_string = params[:search]
      # Use the "includes" method to avoid the "N+1" problem; the "N+1" problem
      # is where we do a separate SQL query for each spconfig to get its groups.
      @spconfigs = Spconfig.includes(:groups).where("entityid LIKE ?",
                                  "%#{search_string}%")
      @breadcrumbs.push(
        [%Q[All SP Configuration(s) (search string "#{search_string}")], ""]
      )
      flash.notice = "Search found #{@spconfigs.size} Service Provider Configuration(s)"
    else
      user       = Utility.get_REMOTE_USER()
      @spconfigs = Spconfig.this_users_spconfigs(user)
      @breadcrumbs.push(
        ["SP Configuration(s) for #{username}", ""]
      )
      flash.notice = "#{@spconfigs.size} Service Provider Configuration(s)"
    end
  end

  # GET /spconfigs/all
  # Only for admins.
  def all
    username = Utility.get_REMOTE_USER()

    if (@is_admin or @is_readonly_admin) then
      @spconfigs = Spconfig.all()
      @breadcrumbs.push(
        ["All SP Configuration(s) (admin)", ""]
      )
      flash.notice = "#{@spconfigs.size} Service Provider Configuration(s)"
      render 'list'
    else
      redirect_to "/spconfigs/list"
    end
  end

  # GET /spconfigs/listincommon
  # Only for admins.
  def listincommon
    username = Utility.get_REMOTE_USER()

    incommon_id = Federation.find_by(name: 'InCommon')

    if (@is_admin or @is_readonly_admin) then
      @spconfigs = Spconfig.where(federation_id: incommon_id)
      @breadcrumbs.push(
        ["All InCommon SP Configuration(s) (admin)", ""]
      )
      flash.notice = "#{@spconfigs.size} Service Provider Configuration(s)"
      render 'list'
    else
      redirect_to "/spconfigs/list"
    end
  end

  # GET /spconfigs/1
  # GET /spconfigs/1.json
  def show
    Utility.logger.info("[controller] SHOW")
    @group_string = @spconfig.workgroup_names.join(', ')

    if (@group_string == '') then
      @group_string = "no owning workgroups"
    end

    @entitlement_string = @spconfig.entitlement_names.join(', ')

    if (@entitlement_string == '') then
      @entitlement_string = "none"
    end

    # Get the certificates
    @certificates = @spconfig.certificates

    breadcrumb = %Q(SP Configuration #{@spconfig.id})
    @breadcrumbs.push(
      [breadcrumb.html_safe, ""]
    )
  end

  # GET /spconfigs/1/attributes
  def attributes
    @resolver_test_uri = URI(Rails.configuration.x.resolver_test.url)
    @username          = Utility.get_REMOTE_USER()

    # We use the form's username_override only if ALL of the following are true:
    #  * the show-attributes-of-others feature is enabled
    #  * the parameter username_override is not empty
    #  * the current logged-in user is an SPDB admin
    if (Rails.configuration.enable_show_attributes_of_other_users) then
      if (@is_admin) then
        @show_override_user = true
        @username_override  = params[:username_override]
        if (not @username_override.blank?()) then
          @username = @username_override
        end
      else
        @show_override_user = false
      end
    end

    @attributes = Utility.get_attributes_for(@spconfig.entityid, @username)

    breadcrumb = %Q(SP Configuration #{@spconfig.id})
    @breadcrumbs.push(
      [breadcrumb.html_safe, spconfig_path(@spconfig.id)]
    )
    breadcrumb = %Q(Attributes)
    @breadcrumbs.push(
      [breadcrumb.html_safe, ""]
    )
  end

  # GET /spconfigs/new
  def new
    if (not @is_writer) then
      redirect_to "/errors/forbidden"
      return
    end

    @spconfig = Spconfig.new
    Utility.logger.info(params.inspect)
    @breadcrumbs.push(
      ["New Service Provider Configuration", ""]
    )
  end

  # GET /spconfigs/1/edit
  def edit
    if (not @is_writer) then
      redirect_to "/errors/forbidden"
      return
    end

    # Does this user control @spconfig? If not, abort.
    @username = Utility.get_REMOTE_USER()
    if (not @spconfig.user_controls?(@username))
      redirect_to "/errors/forbidden_record"
      return
    end

    Utility.logger.info("[controller] EDIT")
    breadcrumb = get_breadcrumb('edit')
    @breadcrumbs.push(
      [breadcrumb, ""]
    )

    if (@spconfig.federation)
      @is_federated = true
    else
      @is_federated = false
    end
    Utility.logger.info('Groups:')
    Utility.logger.info(@spconfig.groups)
    Utility.logger.info('Entitlements:')
    Utility.logger.info(@spconfig.entitlements)
  end

  # POST /spconfigs
  # POST /spconfigs.json
  def create
    if (not @is_writer) then
      redirect_to "/errors/forbidden"
      return
    end

    Utility.logger.info("[controller] CREATE")
    Utility.logger.info("params is #{params.inspect}")
    Utility.logger.info("spconfig_params is #{spconfig_params.inspect}")

    @spconfig = Spconfig.new(spconfig_params)
    @spconfig.owned_by   = Utility.get_REMOTE_USER_or_autoadmin()
    @spconfig.created_by = Utility.get_REMOTE_USER_or_autoadmin()
    @spconfig.updated_by = Utility.get_REMOTE_USER_or_autoadmin()

    # For reasons beyond my understanding, the owning groups attributes
    # from the submitted web form are not avaialable in the model for
    # validation. But we need them to do proper validation, so to get
    # around this we pass them into the model using
    # 'add_groups_attributes', a method we wrote for this purpose.
    #
    # We do the same for entitlements.
    @spconfig.add_groups_attributes(spconfig_params['groups_attributes'])
    @spconfig.add_entitlements_attributes(spconfig_params['entitlements_attributes'])

    if (params['is_federated'] == 'YES') then
      @is_federated = true
      Utility.logger.info("Setting to default federation")
      default_federation = get_default_federation()
      default_federation.spconfigs << @spconfig
    else
      @is_federated = false
      Utility.logger.info("No federation to set")
    end

    respond_to do |format|
      if @spconfig.save
        format.html { redirect_to @spconfig, notice: 'successfully created', alert: '(SP should be available within 15 minutes)' }
        format.json { render :show, status: :created, location: @spconfig }
      else
        format.html { render :new }
        format.json { render json: @spconfig.errors, status: :unprocessable_entity }
      end
    end
  end

  # PATCH/PUT /spconfigs/1
  # PATCH/PUT /spconfigs/1.json
  def update
    if (not @is_writer) then
      redirect_to "/errors/forbidden"
      return
    end

    # Does this user control @spconfig? If not, abort.
    @username = Utility.get_REMOTE_USER()
    if (not @spconfig.user_controls?(@username))
      redirect_to "/errors/forbidden_record"
      return
    end

    @contact_email_old   = @spconfig.contact_email
    @spconfig.attributes = spconfig_params
    @contact_email_new   = @spconfig.contact_email

    @spconfig.updated_by = Utility.get_REMOTE_USER()

    # For reasons beyond my understanding, the owning groups attributes
    # from the submitted web form are not available in the model for
    # validation. But we need them to do proper validation, so to get
    # around this we pass them into the model using
    # 'add_groups_attributes', a method we wrote for this purpose.
    @spconfig.add_groups_attributes(spconfig_params['groups_attributes'])
    @spconfig.add_entitlements_attributes(spconfig_params['entitlements_attributes'])

    # If the "Force save?" checkbox is checked we force the save
    # by updating the updated_at attribute to have the current time.
    updated_at_before_submit = @spconfig.updated_at
    force_save               = spconfig_params['force_save']
    if (force_save.to_i() == 1) then
      @spconfig.update_attribute(:updated_at, Time.now)
    end

    respond_to do |format|
      Utility.logger.info("[update] about to call save method on spconfig")

      if (@spconfig.save()) then
        if (not @is_federated) then
          take_effect = '(Changes should take effect within 15 minutes.)'
        else
          take_effect = ''
        end

        updated_at_after_save = @spconfig.updated_at
        if (updated_at_before_submit != updated_at_after_save) then
          notice = 'SP configuration successfully updated.'
        else
          notice = 'No changes detected.'
          take_effect = ''
        end

        format.html { redirect_to @spconfig, notice: notice, alert: take_effect }
        format.json { render :show, status: :ok, location: @spconfig }
      else
        breadcrumb = get_breadcrumb('edit')
        @breadcrumbs.push(
          [breadcrumb, ""]
        )
        format.html { render :edit }
        format.json { render json: @spconfig.errors, status: :unprocessable_entity }
      end
    end
  end

  # DELETE /spconfigs/1
  # DELETE /spconfigs/1.json
  def destroy
    if (not @is_writer) then
      redirect_to "/errors/forbidden"
      return
    end

    # Does this user control @spconfig? If not, abort.
    @username = Utility.get_REMOTE_USER()
    if (not @spconfig.user_controls?(@username))
      redirect_to "/errors/forbidden_record"
      return
    end

    @contact_email_old = @spconfig.contact_email
    @contact_email_new = nil

    @spconfig.destroy
    respond_to do |format|
      format.html { redirect_to spconfigs_url, notice: 'SP Config was successfully removed.' }
      format.json { head :no_content }
    end
  end

  # GET /spconfigs/entitlementsxml
  def entitlementsxml
    @breadcrumbs.push(
      ["Entitlements XML", ""]
    )
    @entitlements_xml = Spconfig.entitlement_attribute_filter_string()
  end

  def self.log_me(message)
    Utility.logger.info("[controller] #{message}")
  end

  def log_me(message)
    Utility.logger.info("[controller] #{message}")
  end

  private
    # Use callbacks to share common setup or constraints between actions.


    # This function:
    # 1.  Finds the spconfig based on :id
    # 2a. If the InCommon radio button is chosen adds the default federation (InCommon).
    # 2b. If the InCommon radio button is NOT chosen removes the default federation (InCommon) if
    #     necessary.
    def set_spconfig
      log_me("[set_spconfig] finding spconfig with id #{params[:id]}")
      @spconfig = Spconfig.find(params[:id])

      # Deal with the InCommon radio button.
      if (params.key?('is_federated')) then
        default_federation = get_default_federation()
        if (params['is_federated'] == 'YES') then
          log_me("[set_spconfig] the is_federated radio button was selected")

          # This next comand associates the federation do @spconfig but does NOT SAVE
          # this association.
          default_federation.spconfigs << @spconfig
          log_me("[set_spconfig] associated federation '#{default_federation.name}' to spconfig")

          # @spconfig = Spconfig.find(params[:id])
        else
          log_me("[set_spconfig] the is_federated radio button was NOT selected")
          # We may need to remove it. If it is associated, remove it.
          spconfig1 = default_federation.spconfigs.where(id: params[:id])
          if (spconfig1.exists?())
            log_me("[set_spconfig] need to disassociate this record from a federation")
            default_federation.spconfigs.delete(@spconfig)
            @spconfig.federation = nil
            log_me("[set_spconfig] disassociated federation '#{default_federation.name}' from spconfig")
            # @spconfig = Spconfig.find(params[:id])
          else
            log_me("[set_spconfig] NO need to disassociate this record from a federation")
          end
        end
        if (@spconfig.federated?) then
          log_me("[set_spconfig] @spconfig.federated? returned TRUE")
        else
          log_me("[set_spconfig] @spconfig.federated? returned FALSE")
        end
      end

      if (!can_access()) then
        # We allow the show action, but not any other actions
        if ((action_name != 'show') && (action_name != 'attributes')) then
          redirect_to "/errors/forbidden_record"
          return
        end
      end
    end

    # Does the current user (i.e., SUNetID) "own" this configuration?
    # That is, does the current user satisfy one (or more) of the
    # following:
    #
    #   1. user matches owned_by
    #   2. user is an ADMIN
    #   3. user is in one of the owning groups
    #
    # If yes, return true, otherwise return false.
    def can_access()
      user = Utility.get_REMOTE_USER()
      my_owning_workgroups = []
      my_owning_workgroups = Utility.get_my_owning_workgroups(user)

      # 1. user is the owner
      if (@spconfig.owned_by == user) then
        return true
      end

      # 2. user is an SPDB admin.
      if (Utility.admin?(user)) then
        return true
      end

      # 3. user is one (or more) of this models groups.
      if (@spconfig.user_in_groups(user,my_owning_workgroups)) then
        return true
      end

      return false
    end

    def update_partners_list()
      Utility.logger.info("[controller] update partners list")
      # If the action is update or create, add to the partners list.
      if (action_name == 'create') then
        # CREATE
        Utility.add_email_to_partners_mailman_list(@spconfig.contact_email, nil)
      elsif (action_name == 'update') then
        # UPDATE
        # Did the contact_email attribute change?
        if (@contact_email_old != @contact_email_new) then
          Utility.logger.info("contact_email changed: #{@contact_email_old} (old) --> #{@contact_email_new} (new) ")
          Utility.add_email_to_partners_mailman_list(@contact_email_new, @contact_email_old)
        else
          Utility.logger.info("contact_email did NOT change")
        end
      elsif (action_name == 'destroy') then
        # DESTROY
        Utility.logger.info("contact_email in destryed spconfig is #{@contact_email_old}")
        Utility.remove_unused_email_from_partners_list(@contact_email_old)
      end

      return
    end

    def get_total_number_of_configs()
      @total_number_of_configs = Spconfig.count()
    end

    # Never trust parameters from the scary internet, only allow the white list through.
    def spconfig_params
      params.require(:spconfig).permit(:entityid,
                                       :metadata,
                                       :disabled,
                                       :owned_by,
                                       :contact_email,
                                       :login_url,
                                       :comments,
                                       :sign_responses,
                                       :sign_assertions,
                                       :encrypt_assertions,
                                       :is_legacy,
                                       :force_save,
                                       :allow_entitlement_release,
                                       :allow_unlimited_entitlements,
                                       :encryption_exception,
                                       :ignore_valid_until_issues,
                                       :ignore_entityid_format_issues,
                                       groups_attributes: [:id, :workgroup, :_destroy],
                                       entitlements_attributes: [:id, :workgroup, :_destroy],
                                      )
    end

    def get_default_federation
      @default_federation = Federation.find_by name: 'InCommon'
    end

    def set_breadcrumbs_spconfigs()
      @breadcrumbs = Utility.first_level_breadcrumb
      @breadcrumbs.push(['SP Configurations', '/spconfigs/index'])
      @model_name = 'spconfig'
    end

    def get_breadcrumb(action)
      if (action == 'edit') then
        breadcrumb = %Q(Edit SP&nbsp; <span class="sphighlight">#{@spconfig.entityid}</span>)
        return breadcrumb.html_safe
      end
    end

end
