require 'utility'
require 'privgroup'

class Awsaccount < ActiveRecord::Base

  has_many :awsidproviders
  has_many :awsowninggroups
  has_many :awsrolemappings, through: :awsidproviders

  accepts_nested_attributes_for :awsowninggroups,
                                reject_if:     :all_blank,
                                allow_destroy: true
  accepts_nested_attributes_for :awsidproviders,
                                reject_if:     :all_blank,
                                allow_destroy: true

  # We keep track of versions with paper_trail
  has_paper_trail

  ## VALIDATIONS
  validate  :at_least_one_owning_group
  validate  :awsaccount_number_is_valid
  validates :number, presence: {message: 'You must supply a valid account number.'}
  validates_format_of :master,
                      with: /\A\d{12}\Z/,
                      message: 'An account number must be exactly 12 digits.',
                      if: -> { master.present? }
  validates_format_of :nickname,
                      with:    /\A[\w\ \-]{3,64}\Z/,
                      message: 'A nickname can only contain letters, '\
                               'numbers, dashes, and spaces, and must '\
                               'be between 3 and 64 characters long.'
  validate :correct_length

  after_create :after_create_callback
  after_save   :after_save_callback
  before_save  :before_save_callback

  strip_attributes

  # We want a place to put the groups_attributes passed in from the
  # controller.
  attr_accessor :xgroups_attributes

  # A useful constant
  AWSACCOUNTS_RESOLVER_FILE_NAME = 'attribute-resolver-amazon-auto.xml'.freeze

  def provider_type_to_name(type)
    provider_type_to_name_hash = {
      'prod' => 'stanford-idp',
      'uat'  => 'stanford-idp-uat',
    }
    if (not provider_type_to_name_hash.key?(type)) then
      raise "unknown provider type '#{type}"
    else
      return provider_type_to_name_hash[type]
    end
  end

  # Return true if this is a "regular" AWS account
  def regular_aws?()
    return self.instance == 'aws'
  end

  # Return a "pretty" version of the instance:
  #  - "aws"        --> "AWS"
  #  - "aws-us-gov" --> "GovCloud"
  def instance_pretty()
    if (self.instance == 'aws') then
      return 'AWS'
    elsif (self.instance == 'aws-us-gov')
      return 'GovCloud'
    else
      raise "found unknown instance type '#{self.instance}'"
    end
  end

  # Ensure that the account number is valid and not on the
  # blacklist.
  def awsaccount_number_is_valid
    if (number !~ /\A\d{12}\Z/) then
      errors.add(:base, 'An account number must be exactly 12 digits.')
    elsif (Rails.configuration.awsaccounts_black_list.include?(number))
      errors.add(:base, "The account number '#{number}' is not allowed "\
                        'to be managed by this application')
    else
      # Nothing
    end
  end

  ## awsidproviders section
  def saml_providers_formatted()
    arns = []
    awsidproviders.each do |awsidprovider|
      arns << awsidprovider.arn
    end
    return arns.join(', ')
  end

  # If none, return nil
  def get_provider_of_this_type(type)
    name = provider_type_to_name(type)
    awsidproviders.each do |awsidprovider|
      if (awsidprovider.arn =~ /saml-provider\/#{name}\z/) then
        return awsidprovider
      end
    end
    return nil
  end

  def get_prod_provider()
    return self.get_provider_of_this_type('prod')
  end

  # Returns either 'aws' or 'aws-us-gov'
  def self.arn_to_instance(arn_string)
    if    (arn_string =~ %r{^arn:aws:}) then
      return 'aws'
    elsif (arn_string =~ %r{^arn:aws-us-gov:}) then
      return 'aws-us-gov'
    else
      raise "unrecognized arn: #{arn_string}"
    end
  end

  # AWS provider ARN's look like "arn:aws:iam::548206471889:saml-provider/stanford-idp"
  def has_idp_prod_provider?()
    return has_idp_this_provider_type?('prod')
  end

  def has_idp_uat_provider?()
    return has_idp_this_provider_type?('uat')
  end

  # Has the provider type passed. Must be one of 'prod' or 'uat'.
  def has_idp_this_provider_type?(type)
    name = provider_type_to_name(type)

    awsidproviders.each do |awsidprovider|
      if (awsidprovider.arn =~ /saml-provider\/#{name}\z/) then
        return true
      end
    end

    # If we get here, no matches.
    return false
  end

  # Add or delete the provider
  def update_provider(provider_type: type, add: true, instance: self.instance)
    Utility.logger.info("instance is #{instance}")
    if (add) then
      # ADD
      if (has_idp_this_provider_type?(provider_type)) then
        # Nothing to do.
        return
      else
        # Need to add it
        awsidprovider = Awsidprovider.new()
        name = provider_type_to_name(provider_type)
        arn = "arn:#{instance}:iam::#{number}:saml-provider/#{name}"
        awsidprovider.arn = arn
        self.awsidproviders << awsidprovider
        Utility.logger.info("awsidprovider is #{awsidprovider.inspect}")
      end
    else
      # REMOVE
      if (has_idp_this_provider_type?(provider_type)) then
        Utility.logger.info("has provider of type #{provider_type}")
        # Remove it
        awsidprovider_to_remove = self.get_provider_of_this_type(provider_type)
        if (!awsidprovider_to_remove.nil?) then
          Utility.logger.info('found awsidprovider; now I will destroy it')
          self.awsidproviders.destroy(awsidprovider_to_remove)
        end
      else
        # Nothing to do.
      end
    end
  end

  def remove_prod_provider(instance: nil)
    self.update_provider(provider_type: 'prod', add: false, instance: instance)
  end

  def add_prod_provider(instance: nil)
    self.update_provider(provider_type: 'prod', add: true, instance: instance)
  end

  # Return an array of the workgroup names. Will return the empty
  # array if there are none.
  def workgroup_names()
    if (awsowninggroups.nil?() or awsowninggroups.empty?())
      return []
    else
      return awsowninggroups.map { |g| g.workgroup }
    end
  end

  # Function to add a parameter (used to pass information from controller
  # to model).
  def add_groups_attributes(groups_attributes)
    if (not groups_attributes.nil?) then
      self.xgroups_attributes = groups_attributes
    else
      self.xgroups_attributes = {}
    end
    return
  end

  def at_least_one_owning_group()
    # If remote_user is 'auto-migration' skip this validation.
    if ($remote_user == 'auto-migration') then
      return
    end

    groups_attributes = self.xgroups_attributes
    Utility.logger.info "xgroups_attributes is #{xgroups_attributes.inspect}"
    groups_attributes.each_value do |group_attr|
      # group_attr is a hash. At least one must have its key
      # '_destroy' not be 1.
      if (group_attr.key?('_destroy')) then
        destroy_value = group_attr['_destroy']
        if (destroy_value != '1') then
          # At least one is not to be destroyed, so we are
          # good.
          return
        end
      end
    end

    # If we get here, either there were no owning groups, or else there
    # were owning groups but they were all to be destroyed. In either
    # case, that is a problem.
    errors.add(:base, 'Must have at least one owning group')
    return
  end

  def correct_length()
    nickname = self.nickname
    if (!nickname) then
      errors.add(:base, 'An AWS account nickname must be provided.')
    elsif (nickname.length > 64) then
      errors.add(:base, 'An AWS account nickname must be no longer than 64 characters.')
    end
  end

  def saml_arn()
    if (self.number.blank?()) then
      return ''
    else
      return "arn:#{self.instance}:iam::#{number}:saml-provider/stanford-idp"
    end
  end

  def to_s()
    rv = <<~EOS
      instance:  #{instance}
      number:    #{number}
    EOS

    # Add the id providers
    idproviders = self.awsidproviders

    providers_formatted = ''
    idproviders.each do |idprovider|
      mappings = idprovider.awsrolemappings
      mapping_formatted = ''
      mappings.each do |mapping|
        mapping_formatted += "  workgroup: #{mapping.workgroup}\n  role: #{mapping.role}\n"
      end

      if (mapping_formatted == '') then
        mapping_formatted = '  NONE'
      end

      rv += <<~EOS
        awsidprovider: #{idprovider.arn}
        #{mapping_formatted.rstrip()}
      EOS
    end

    return rv
  end

  # Return true if user is in (or is an administrator of) at least one of
  # this object's groups
  def user_in_workgroups(user)
    awsowninggroups.each do |group|
      Utility.logger.info "checking to see if #{user} is a member of administrator of the group #{group.workgroup}"
      privgroup = Privgroup.new(group.workgroup)
      if (privgroup.user_is_member?(user))
        return true
      end
    end

    return false
  end

  # Return an array of the awsaccounts that the given user has access
  # to. That is all the awsaccounts that contain a group that the user is
  # a member of.
  def self.this_users_awsaccounts(user)
    # Find all awsaccounts that have an associated group
    # containing user as a member.
    awsaccounts_group_owned = []

    awsaccounts_all = Awsaccount.all()
    awsaccounts_all.each do |awsaccount|
      if (awsaccount.user_in_workgroups(user)) then
        awsaccounts_group_owned.push(awsaccount)
      end
    end

    return awsaccounts_group_owned
  end

  ###
  ### Methods relating to the awsaccounts auto file
  ###

  # Returns the full path to the awsaccounts auto file.
  def self.awsaccounts_auto_file_path()
    local_basepath = Rails.configuration.awsaccounts_git_local_path
    local_name     = Rails.configuration.awsaccounts_git_local_name

    full_path = File.join(local_basepath, local_name, AWSACCOUNTS_RESOLVER_FILE_NAME)
    return full_path
  end

  def self.awsaccounts_auto_git_dir()
    local_basepath = Rails.configuration.awsaccounts_git_local_path
    local_name     = Rails.configuration.awsaccounts_git_local_name

    git_dir = File.join(local_basepath, local_name)
    return git_dir
  end

  # Write the awsaccounts auto file. We order first by the AWS account
  # number, then by the workgroup name. This function only writes the
  # file, it does not commit or push it.
  def self.write_awsaccounts_auto_file()
    # Step 1. Setup output file
    filepath = Awsaccount.awsaccounts_auto_file_path()

    f = File.new(filepath, 'w')

    # Step 2. Write overall file header
    f.write(Awsrolemapping.rolemapping_global_header())

    # Step 3. Write each instance
    f.write(Awsaccount.write_awsaccounts_instance('aws'))
    f.write(Awsaccount.write_awsaccounts_instance('aws-us-gov'))

    # Step 4. Write global footer.
    f.write(Awsrolemapping.rolemapping_global_footer())

    # Step 5. Close file and return
    f.close()

    return
  end

  # Write the awsaccounts mappings for all the awsaccounts of the given
  # instance. Currently, we support only these two instances:
  #
  #  - aws
  #  - aws-us-gov
  #
  # The header files for each instance are defined in the function
  # awsrolemapping_instance_header.
  #
  # We order first by the AWS account number, then by the workgroup name.
  #
  # This function takes as parameter the instance name and writes to a new temporary file.
  #
  # Returns a string with all the mappings
  def self.write_awsaccounts_instance(instance_name)
    # Step 1. Get the header and footer.
    header = Awsrolemapping.rolemapping_header(instance_name)
    footer = Awsrolemapping.rolemapping_footer(instance_name)

    # Step 2. Get the mappings.

    # Indent the ValueMaps this much.
    indent = ' ' * 8

    mapping_string = ''
    Awsaccount.all().order(:number).each do |awsaccount|
      if (awsaccount.instance != instance_name)
        Utility.logger.info("skipping awsaccount #{awsaccount.number} has instance #{awsaccount.instance} " \
                            "that does not match #{instance_name}")
        next
      end

      puts "processing account #{awsaccount.number}"
      awsaccount.awsidproviders.each do |awsidprovider|
        # We only process id providers that are "production" stanford
        # SAML idp providers
        if (awsidprovider.regular_prod?()) then
          # Put in a comment with the accounts number and nickname
          mapping_string += awsaccount.account_nickname_header(indent)

          # This is a regular production type, so add its mappings.

          mappings = awsidprovider.awsrolemappings
          mappings_sorted = mappings.sort { |a, b| a.workgroup.downcase <=> b.workgroup.downcase }
          mappings_sorted.each do |mapping|
            # Remember: each mapping is actually an Awsrolemapping object.
            mapping_string += mapping.to_value_map(indent)
          end
        else
          # Skipping this one
          puts "  skipping id provider #{awsidprovider.arn}"
        end
      end
    end

    # If there are any, put them all together.
    if (mapping_string != '') then
      result = header + mapping_string + footer
    else
      result = nil
    end

    return result
  end

  def account_nickname_header(indent)
    if (nickname.blank?()) then
      nickname_formatted = 'no nickname'
    else
      nickname_formatted = nickname
    end

    header = "\n#{indent}<!-- AWS account #{number} (#{nickname_formatted}) -->\n"
    return header
  end

  # If the awsaccounts auto file has changed, write it to the git
  # directory, commit, and push,
  def self.save_awsaccounts_auto_file(force_update = false)
    # First of all, do we even _need_ to write the metadata file?
    if (not Awsrolemapping.should_awsaccounts_auto_file_be_exported(force_update)) then
      # Utility.logger.info "no need to update awsaccounts auto file at this time"
      # return
    else
      Utility.logger.info 'awsaccounts auto file needs to be updated'
    end

    ## If we get here then we need to write the file.

    # Step 1. Write the file.
    Awsaccount.write_awsaccounts_auto_file()

    # Step 2. Commit and push the change.
    git_dir = Awsaccount.awsaccounts_auto_git_dir()
    message = "auto-commit for awsaccounts file '#{AWSACCOUNTS_RESOLVER_FILE_NAME}'"
    Utility.git_commit_and_push(git_dir, message)

    return
  end

  # From the passed in array remove all the awsaccounts that the current
  # user does not have access to
  def self.return_only_accessible(awsaccounts, username)
    if (Utility.admin?(username)) then
      return awsaccounts
    end

    accessible_awsaccounts = []
    awsaccounts.each do |awsaccount|
      # username is one (or more) of this models groups.
      if (awsaccount.user_in_workgroups(username)) then
        accessible_awsaccounts.push(awsaccounts)
      end
    end
    return accessible_awsaccounts
  end

  private

  def after_create_callback()
    # After create, add the 'stanford-idp' production provider
    self.add_prod_provider()
    Awsrolemapping.touch()
  end

  def before_save_callback
    # 1. If the instance changed, delete the old idprovider and add the
    # new one.
    if (not self.instance_change.nil?) then
      instance_old = self.instance_was
      instance_new = self.instance
      Utility.logger.info "[before_save_callback] change in 'instance' "\
                          "field detected: old: #{instance_old}, new: #{instance_new}"
      Utility.logger.info "removing awsidprovider for old instance #{instance_old}"
      self.remove_prod_provider(instance: instance_old)
      Utility.logger.info "adding awsidprovider for new instance #{instance_new}"
      self.add_prod_provider(instance: instance_new)
    end
  end

  def after_save_callback
    # 1. Remove any duplicate workgroups.
    workgroup_hash = {}
    self.awsowninggroups.each do |awsowninggroup|
      workgroup = awsowninggroup.workgroup
      if (workgroup_hash.key?(workgroup)) then
        # This is already in the list, so delete
        awsowninggroup.delete()
        Utility.logger.info("deleting duplicate awsowninggroup '#{workgroup}'")
      else
        workgroup_hash[workgroup] = true
      end
    end

    # 2. Indicate that there is a change that warrants regenerating the
    # role-mapping file.
    Utility.logger.info 'awsaccounts change detected; '\
                        'touching the updatedtimes table'
    Awsrolemapping.touch()
  end
end
