require 'utility'
require 'privgroup'

class Awsrolemapping < ActiveRecord::Base
  belongs_to :awsidprovider

  # We keep track of versions with paper_trail
  has_paper_trail

  ## VALIDATIONS
  validate  :group_is_valid
  validate  :role_name_is_valid

  after_destroy :after_destroy_callback
  after_save    :after_save_callback

  # In case we need to skip validating that the workgroup is in
  # LDAP (e.g., when doing an initial load), set this attribute
  # to true.
  attr_accessor :skip_workgroup_validation

  # The label we use in the Updatedtime class specific to awsrolemapping
  # changes.
  @@awsrolemapping_updated_label = 'awsrolemapping-last-changed'
  def self.awsrolemapping_updated_label()
    return @@awsrolemapping_updated_label
  end

  # Validate that the workgroup is a real group in the OpenLDAP groups tree.
  def group_is_valid
    Utility.logger.debug("[awsrolemapping] checking that group #{workgroup} is in OpenLDAP")
    if (skip_workgroup_validation) then
      Utility.logger.debug("[awsrolemapping] skipping validation")
    else
      if (Privgroup.new(workgroup).valid?()) then
        Utility.logger.debug("[awsrolemapping] good news: #{workgroup} is in OpenLDAP")
        return
      else
        Utility.logger.info("[awsrolemapping] #{workgroup} is NOT in OpenLDAP")
        errors.add(:workgroup, "workgroup '#{workgroup}' not a valid workgroup.")
      end
    end
  end
  # A submitted role name can be either a full ARN or just the role name.
  def role_name_is_valid
    instance = self.awsidprovider.awsaccount.instance
    number   = self.awsidprovider.awsaccount.number

    if (mtch = /\Aarn:#{instance}:iam::#{number}:role\/([\w\-\.]+)\Z/.match(role)) then
      rolename = mtch[1]
      Utility.logger.debug("[awsrolemapping] role string appears to be a full ARN (#{rolename}); that's OK")

      # Reset role to be just the rolename
      self.role = rolename
    elsif (role =~ /\A[\w\-\.]+\Z/) then
      Utility.logger.debug("[awsrolemapping] role string appears to be a bare role name; that's OK")
    else
      # If we get here, not so great.
      errors.add(:base, "The role name '#{role}' is not valid: for 'AWS IAM Role' " +
                        "enter either the full ARN or just the role name; the role name " +
                        "can only contain letters, numbers, dashes, and underscores")
    end
  end

  # Return the full role arn
  def role_arn
    instance = self.awsidprovider.awsaccount.instance
    number   = self.awsidprovider.awsaccount.number

    return "arn:#{instance}:iam::#{number}:role/#{role}"
  end

  def to_value_map(indent = '')
    # Get the id provider
    idprovider_arn = self.awsidprovider.arn
    number         = self.awsidprovider.awsaccount.number
    <<eos
#{indent}<!-- #{number}: workgroup "#{workgroup}" maps to role "#{role}"   -->
#{indent}<ValueMap>
#{indent}    <ReturnValue>#{idprovider_arn},#{self.role_arn}</ReturnValue>
#{indent}    <SourceValue>#{workgroup}</SourceValue>
#{indent}</ValueMap>

eos
  end

  def self.rolemapping_global_header()
    return<<eoh
<AttributeResolver xmlns="urn:mace:shibboleth:2.0:resolver" xmlns:resolver="urn:mace:shibboleth:2.0:resolver"
                   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:pc="urn:mace:shibboleth:2.0:resolver:pc"
                   xmlns:ad="urn:mace:shibboleth:2.0:resolver:ad" xmlns:dc="urn:mace:shibboleth:2.0:resolver:dc"
                   xmlns:enc="urn:mace:shibboleth:2.0:attribute:encoder" xmlns:sec="urn:mace:shibboleth:2.0:security"
                   xsi:schemaLocation="urn:mace:shibboleth:2.0:resolver classpath:/schema/shibboleth-2.0-attribute-resolver.xsd
                                       urn:mace:shibboleth:2.0:resolver:pc classpath:/schema/shibboleth-2.0-attribute-resolver-pc.xsd
                                       urn:mace:shibboleth:2.0:resolver:ad classpath:/schema/shibboleth-2.0-attribute-resolver-ad.xsd
                                       urn:mace:shibboleth:2.0:resolver:dc classpath:/schema/shibboleth-2.0-attribute-resolver-dc.xsd
                                       urn:mace:shibboleth:2.0:attribute:encoder classpath:/schema/shibboleth-2.0-attribute-encoder.xsd
                                       urn:mace:shibboleth:2.0:security classpath:/schema/shibboleth-2.0-security.xsd">

    <!--
         DO NOT EDIT THIS FILE! DO NOT EDIT THIS FILE! DO NOT EDIT THIS FILE! DO NOT EDIT THIS FILE!

         This file is generated automatically.

         DO NOT EDIT THIS FILE! DO NOT EDIT THIS FILE! DO NOT EDIT THIS FILE! DO NOT EDIT THIS FILE!
    -->

    <!-- ========================================== -->
    <!--      Attribute Definitions                 -->
    <!-- ========================================== -->

    <!-- awsname is used for Amazon authentication -->
    <AttributeDefinition xsi:type="Simple"
                         id="awsname"
                         activationConditionRef="ac_AttributeResolver_Amazon">
        <InputAttributeDefinition ref="uid" />
        <AttributeEncoder
            xsi:type="SAML2String"
            name="https://aws.amazon.com/SAML/Attributes/RoleSessionName"
            friendlyName="RoleSessionName" />
    </AttributeDefinition>

eoh

  end

  def self.rolemapping_global_footer()
    return<<eof
</AttributeResolver>
eof
  end


  def self.rolemapping_header(instance_name)
    headers = {}

    headers['aws'] = <<eoh
    <!-- AWS (regular cloud) ONLY   -->
    <AttributeDefinition xsi:type="Mapped"
                             id="awsrole"
                             activationConditionRef="ac_AttributeResolver_Amazon">
        <InputDataConnector ref="suLDAP" attributeNames="suPrivilegeGroup" />
        <AttributeEncoder
            xsi:type="SAML2String" name="https://aws.amazon.com/SAML/Attributes/Role"
            friendlyName="Role" />
eoh

    headers['aws-us-gov'] = <<EOK
    <!-- AWS GOVCLOUD ONLY      -->
    <AttributeDefinition xsi:type="Mapped"
                             id="awsgovrole"
                             activationConditionRef="ac_AttributeResolver_Amazon">
        <InputDataConnector ref="suLDAP" attributeNames="suPrivilegeGroup" />
        <AttributeEncoder
            xsi:type="SAML2String" name="https://aws.amazon.com/SAML/Attributes/Role"
            friendlyName="Role" />
EOK

    return headers[instance_name]
  end

  def self.rolemapping_footer(instance_name)
    instance_name_formatted = instance_name.upcase()

    return<<EOF
        <!--  END OF MAPPINGS FOR #{instance_name_formatted}  -->
    </AttributeDefinition>



EOF
  end

  def self.should_awsaccounts_auto_file_be_exported(force_update = false)
    filepath = Awsaccount.awsaccounts_auto_file_path()

    # We force a file write if the file is more than 5 minutes old.
    return Utility.should_file_be_exported(
             filepath:           filepath,
             updatedtime_name:   @@awsrolemapping_updated_label,
             force_update:       force_update,
             time_limit_seconds: 5 * 60 * 60,
             action_label:       'awsaccounts-auto-file-changes'
           )
  end

  # Call this method when there is any change that requires a new awsmapping file.
  def self.touch()
    Updatedtime.touch(@@awsrolemapping_updated_label)
  end

  # Touch the updatedtimes table if there are any changes.
  def after_save_callback
    Awsrolemapping.touch()
  end

  def after_destroy_callback
    Awsrolemapping.touch()
  end

end
