require 'utility'
require 'privgroup'

class Entitlement < ActiveRecord::Base
  belongs_to :spconfig

  validate :entitlement_is_valid
  validate :entitlement_is_not_private

  strip_attributes

  ## CALLBACKS
  after_create  :after_create_callback
  after_destroy :after_destroy_callback

  # The label we use in the Updatedtimes table so that we can track
  # if we need to re-generate the entitlements attribute-filter XML file
  UPDATED_TIME_ENTITLEMENTS_XML_FILE = 'spconfig-entitlements-last-changed'.freeze

  # Validate that the workgroup is a real group in the OpenLDAP groups tree.
  def entitlement_is_valid
    Utility.logger.debug("checking that entitlement #{workgroup} is in OpenLDAP")
    if (Privgroup.new(workgroup).valid?()) then
      Utility.logger.debug("good news: #{workgroup} is in OpenLDAP")
      return
    else
      Utility.logger.info("#{workgroup} is NOT in OpenLDAP")
      errors.add(:workgroup, "Entitlement '#{workgroup}' not a valid workgroup.")
    end
  end

  # If a workgroup is marked private in the LDAP we do not allow it to be released
  # via the SPDB web interface mechanism.
  def entitlement_is_not_private
    Utility.logger.debug("checking that entitlement #{workgroup} is not private")
    if (Privgroup.new(workgroup).private?()) then
      Utility.logger.info("#{workgroup} is private")
      errors.add(:workgroup, "Entitlement '#{workgroup}' is a private workgroup; "\
                             "unable to release membership.")
      return
    else
      Utility.logger.info("good news: #{workgroup} membership allowed to be released")
    end
  end

  def self.log_me(message)
    Utility.logger.info "[entitlement model] #{message}"
  end

  def log_me(message)
    Utility.logger.info "[entitlement model] #{message}"
  end

  def after_create_callback
    log_me("adding entitlement #{workgroup} belonging to spconfig #{self.spconfig.id}")
    Entitlement.touch()
  end

  def after_destroy_callback
    log_me("removing entitlement #{workgroup} belonging to spconfig #{self.spconfig.id}")
    Entitlement.touch()
  end

  # Call this method when there is any change that requires a new entitlements
  # attribute-filter XML file
  def self.touch()
    Updatedtime.touch(UPDATED_TIME_ENTITLEMENTS_XML_FILE)
  end

  def self.entitlements_file_filepath()
    basedir = Rails.configuration.sp_entitlements_xml_file_basedir
    return  File.join(basedir, 'attribute-filter-entitlements-auto.xml')
  end

  # Note that force_update is trumped by the
  # "auto-entitlements-enabled" administrative configuration. This is
  # a bit confusing, but keep it in mind.
  def self.should_entitlements_file_be_exported(force_update = false)
    if (AdminConfiguration.ae_enabled?()) then
      return Utility.should_file_be_exported(
               filepath:           Entitlement.entitlements_file_filepath(),
               updatedtime_name:   UPDATED_TIME_ENTITLEMENTS_XML_FILE,
               force_update:       force_update,
               time_limit_seconds: 2 * 60 * 60,
               action_label:       'entitlement-changes'
             )
    else
      log_me("auto-entitlements-enabled administrative configuration is false, so not exporting entitlements file")
      return false
    end
  end
end
