require 'fileutils'
require 'net/http'
require 'net/https'
require 'uri'
require 'json'
require 'utility'
require 'digest'
require 'yaml'

require 'privgroup'

class Spconfig < ActiveRecord::Base
  # Each spconfig can zero, one, or many owning groups, but we don't want
  # duplicates.
  has_many :groups

  # Each spconfig can zero, one, or many entitlements (i.e., Stanford
  # Workgroups), but we don't want duplicates.
  has_many :entitlements

  belongs_to :federation

  # Each spconfig can zero, one, or many certificates. We want those
  # certificates to go away when the spconfig goes away.
  has_many :certificates, dependent: :destroy

  accepts_nested_attributes_for :groups,
                                reject_if:     :all_blank,
                                allow_destroy: true

  accepts_nested_attributes_for :entitlements,
                                reject_if:     :all_blank,
                                allow_destroy: true

  strip_attributes

  # We want a place to put the groups_attributes passed in from the
  # controller. Similarly for entitlements.
  attr_accessor :xgroups_attributes
  attr_accessor :xentitlements_attributes

  # We want to give admins the (hopefully rarely used) option to supress
  # the validity checks certain properties of the SP config.
  # attr_accessor :ignore_cert_issues
  attr_accessor :ignore_valid_until_issues
  attr_accessor :ignore_entityid_format_issues

  # Force saving the record regardless of changes.
  attr_accessor :force_save

  # We keep track of versions with paper_trail
  has_paper_trail

  ## VALIDATIONS
  validate  :metadata_is_valid
  validate  :at_least_one_owning_group
  validate  :does_not_exceed_maximum_entitlements
  validates :contact_email, presence: {message: 'You must supply a valid '\
                                                'mailing list address',}
  validates :entityid, uniqueness: {message: 'This entityID is already being '\
                                             'used by another configuration',
                                    case_sensitive: false,},
                       if: -> { entityid.present? }

  ## CALLBACKS
  before_validation :get_validation_results

  before_destroy do
    # Uncomment the following to delete all versions of an spconfig object
    # when deleting the spconfig itself. Of course, doing that means you would
    # not be able to recover a deleted spconfig object from the versions table.
    # That is why we leave it commented.

    # self.versions.destroy_all
  end

  # We want to mark for export the metadata files after any create or delete.
  after_create  :after_create_callback
  after_destroy :after_destroy_callback

  # We want to strip out any Signatures before saving.
  before_save   :before_save_callback

  # We want to mark for export the metadata file after any save where there is
  # a metadata change.
  after_save    :after_save_callback

  ####    ####  ####    ####  ####    ####  ####    ####  ####    ####  ####    ####

  # Some constants

  # Here is where we define the mapping from metadata type to the basename
  # of its output file.
  METADATA_TYPE_TO_BASENAME = {
    'combined'   => 'metadata',
    'legacy'     => 'metadata-legacy',
    'non-legacy' => 'metadata-blanket',
  }.freeze

  def self.metadata_type_to_basename()
    return METADATA_TYPE_TO_BASENAME
  end

  # Each metadata file has a 'Name' parameter in the header and this name
  # must be unique for each metdata file consumed by the IdP. Thus, we
  # define a unique header name for each metadata file we output.
  METADATA_TYPE_TO_HEADERNAME = {
    'combined'   => 'urn:mace:stanford.edu:shibboleth:stanford',
    'legacy'     => 'https://spdb.stanford.edu/spmetadata/metadata-legacy.xml',
    'non-legacy' => 'https://spdb.stanford.edu/spmetadata/metadata-blanket.xml',
  }.freeze

  # A few useful instance variables
  @metadata_validation_results = nil
  @sp_info                     = nil
  @saml_validation_errors      = nil

  def federated?()
    if (self.federation) then
      return true
    else
      return false
    end
  end

  def self.log_me(message)
    Utility.logger.info "[model] #{message}"
  end

  def log_me(message)
    Utility.logger.info "[model] #{message}"
  end

  def metadata_is_valid
    log_me('entering metadata_is_valid')

    # Go through the validation errors adding them as appropriate.
    if (!@saml_validation_errors.nil?)
      if (@saml_validation_errors.key?('200')) then
        # If the metadata contains '<?xml' skip the other errors.
        errors.add(:base, %q(Metadata must NOT contain the string '<?xml'))
        return
      elsif (@saml_validation_errors.key?('300')) then
        # If the metadata is not valid XML skip the other errors.
        errors.add(:base, 'Metadata is not valid XML')
        return
      elsif (@saml_validation_errors.key?('350')) then
        # If the metadata is not valid SAML XML skip the other errors.
        error_value = @saml_validation_errors['350'][0]
        error_value = error_value.strip()

        # Remove everything before the first colon (this gets rid of the
        # extra 'Metadata is not valid SAML 2.0 XML'.
        error_value.gsub!(/^[^:]*:/, '')

        errors.add(:base, "350: #{error_value}")
        return
      elsif (@saml_validation_errors.key?('370')) then
        # Contains a non-parsable certificate.
        errors.add(:base, 'Metadata contains a non-parsable certificate')
        return
      else
        # For the other errors, add them now. Note that we skip certificate
        # errors (5xx) and deal with them below.
        # Sort the errors.
        error_numbers = @saml_validation_errors.keys.sort
        error_numbers.each do |error_number|
          if (error_number =~ /^5/) then
            # This is a certificate error, so skip (for now).
            next
          end
          if (error_number == '410') then
            # This is a validUntil error we which we handle ourselves
            # below, so skip (for now).
            next
          end
          error_texts = @saml_validation_errors[error_number]
          error_texts.each do |error_text|
            if (error_number == '610') then
              # This is an empty element error.
              if (match = error_text.match(/contains a (\S+) element that is empty but should not be/))
                element = match.captures[0]
                errors.add(:base, "The element '#{element}' cannot be empty")
              end
              next
            else
              errors.add(:base, "Metadata [#{error_number}]#{error_text}")
            end
          end
        end
      end
    end

    # If the entityID is not defined, stop since other validation tests depend on
    # there being an entityID.
    if (not self.entityid) then
      errors.add(:base, 'Missing entityId.')
      return
    end

    # Deal with federated entries first.
    if (self.federated?) then
      log_me('this SP is part of an external federation')
      # Check that this entityId is in the incommon metdata. If not, add
      # error.
      if (not Utility.is_an_incommon_entity_id(self.entityid)) then
        errors.add(:base, "This entityId is not in my #{self.federation.name} metadata.")
      end

      log_me('has an associated federation so skipping metadata validation')
      # Since there is an associated federation, we skip all metadata validation.
      self.metadata = "SP metadata provided by the #{self.federation.name} federation"
      return
    else
      log_me('does not have an associated federation')
    end

    # Is the metadata associated with a federation?
    if (self.metadata =~ /SP metadata provided by the .* federation/) then
      log_me('metadata is from an external federation')
      metadata_from_external_federation = true
    else
      metadata_from_external_federation = false
    end

    if (self.force_save.to_i() == 1) then
      force_metadata_save = true
    else
      force_metadata_save = false
    end

    # Skip validation of the metadata if this is an old record AND the
    # metadata field has NOT changed and this metadata is not from an
    # external federation.
    if (    !self.metadata_was.nil?   \
        and !self.metadata_changed?   \
        and !force_metadata_save \
        and !metadata_from_external_federation) then
      log_me('skipping validation as this is an old record, ' \
             'the metadata has not changed, we are not forcing the save, ' \
             'and metadata is not from an external federation')
      return
    end

    # If the user has indicated that this SP is part of the InCommon Federation,
    # set metatadata to some default string and skip the rest of the validation.
    if (self.metadata.nil? or self.metadata.empty?) then
      errors.add(:base, 'Metadata cannot be empty.')
      return
    end

    # If the whitespace-trimmed version of the metadata from BEFORE
    # the edit agrees with the whitespace-trimmed version of the
    # metadata from AFTER the edit, then there is no reason to
    # validate (unless the old metadata was the string we put in for
    # InCommon records). Unless we are forcing the save.
    log_me('check to see if metadata has changed')
    if (!force_metadata_save and self.metadata_was and
        (Utility.normalize_whitespace(self.metadata_was) == Utility.normalize_whitespace(self.metadata)) and
        (!metadata_from_external_federation)
       ) then
      return
    end

    length = @metadata_validation_results.length()

    if (length != 2) then
      raise 'metadata validator failed'
    end

    # Set the ignore_certificate_expiration and ignore_no_certificates
    # booleans based on the encryption_exceptions value.
    if ((not Rails.configuration.check_cert_expirations) or
        self.encryption_exception) then
      ignore_certificate_expiration = true
    else
      ignore_certificate_expiration = false
    end

    if ((not Rails.configuration.check_for_encrypting_cert) or
        self.encryption_exception) then
      ignore_no_encrypting_certificates = true
    else
      ignore_no_encrypting_certificates = false
    end

    if (self.ignore_entityid_format_issues.to_i() == 1) then
      ignore_entityid_format_issues = true
    else
      ignore_entityid_format_issues = false
    end

    # Set the ignore_valid_until_expiration boolean based on the
    # global Rails configuration and checkbox.
    if ((not Rails.configuration.check_valid_until) or
        (self.ignore_valid_until_issues.to_i() == 1)) then
      ignore_valid_until_expiration = true
    else
      ignore_valid_until_expiration = false
    end

    ## ## INCOMMON METADATA
    # Don't allow submission of metadata using an InCommon entityID (may
    # have to change this later).
    if (Utility.is_an_incommon_entity_id(self.entityid)) then
      msg = <<~MESSAGE.strip
        The entityID supplied ('#{self.entityid}')
        is an InCommon entityID; this SP's metadata
        is already known to our IdPs.
      MESSAGE
      errors.add(:base, msg)
    end

    ## ## CERTIFICATES

    # We need the certificates for the next couple of sectins.
    certificates = Utility.get_certificate_objects_from_metadata(self.metadata)

    # Get the counting hashes.
    number_certs_of_type = get_cert_use_count(certificates)

    log_me("\# certificates (encryption):   #{number_certs_of_type['encryption']}")
    log_me("\# certificates (signing):      #{number_certs_of_type['signing']}")
    log_me("\# certificates (unspecified):  #{number_certs_of_type['unspecified']}")
    log_me("\# certificates (unrecognized): #{number_certs_of_type['unrecognized']}")

    # The number of certificates suitable for encryption is the sum of the
    # number of certificates of type 'encryption' and type 'unspecified'
    # (remember that a certificate of type 'unspecified' can be used for
    # any purpose).
    number_certs_suitable_for_encryption = number_certs_of_type['encryption'] + \
                                           number_certs_of_type['unspecified']

    ## ## ENCRYPTING CERTIFICATE
    #
    # Unless we are ignoring the whole issue of encrypting certificates,
    # this validation checks that at least one of the certificates can be
    # used for encryption. Note that this validation does not care whether
    # the certificate has expired or not.
    if (not ignore_no_encrypting_certificates) then

      if (number_certs_of_type['unrecognized'].positive?) then
        errors.add(:base, 'One of your certificates has an unrecognized use.')
      end

      exception_url = 'https://uit.stanford.edu/service/saml/exception'
      missing_cert_text = <<~MISSING_CERT_MESSAGE.strip
        Metadata must contain at least one certificate for encryption; if
        your application does not support encrypted responses you must file an
        exception; for more information see <a target="_blank" href="#{exception_url}">
        SAML Encryption Exception Process</a>.
      MISSING_CERT_MESSAGE
      if (certificates.empty? or number_certs_suitable_for_encryption.zero?) then
        errors.add(:base, missing_cert_text)
      end
    end

    # These next variables are used in a couple of the following sections
    # that check for various kinds of expirations. Because some SP's
    # generate a certificate for exactly one year, we can get submissions
    # that are a few hours short of one year. So, set our limit to be one
    # year minus 14 days.
    almost_one_year_of_seconds = 60 * 60 * 24 * (365 - 14)
    now_utc                    = Time.now.utc
    expiration_limit_utc_epoch = now_utc.to_i() + almost_one_year_of_seconds

    ## ## CERTIFICATE EXPIRATION
    #
    # Unless we are ignoring the whole issue of expired certificates, this
    # validation checks that at least one of the _encrypting_ certificates
    # is not expired or too close to expiring. If there _are_ no
    # certificates suitable for encrypting, we skip this validation.
    if ((not ignore_certificate_expiration) and \
        (not number_certs_suitable_for_encryption.zero?)) then

      # This is a hash mapping use type to the number of certificates
      # of that type that are not expired. Note that if no certificates
      # of that use type appear, there will be no key for that use type.
      suitable_for_encryption_count = 0

      # Check certificate expirations. For each KeyDescriptor, at least one
      # certificate must be valid for a year.
      certificates.each do |certificate|
        if (certificate.can_be_used_for_encryption?()) then
          # Is this certificate expiring far enough into the future?
          # NOT DONE HERE! MAKE SURE COMPARISON WORKS
          if (expiration_limit_utc_epoch < certificate.expires_at.utc.to_i) then
            # This certificate is OK, so we increment the counter that keeps track of
            # how many certificates of this usage type are still withing the expiration
            # window.
            suitable_for_encryption_count += 1
          else
            # This certificate is expired or is expiring too soon.
          end
        end
      end

      # Did we find one?
      if (suitable_for_encryption_count.zero?) then
        errors.add(:base,
                   'No certificate found with at least ' \
                   "one year left before expiration (use: #{usage_type})")
      end
    end

    ## ## METADATA: strip signatures
    begin
      stripped_string = Utility.strip_signatures(self.metadata)
    rescue
      log_me('error stripping signature')
      errors.add(:base, 'Error stripping signatures; maybe XML is bad?')
    else
      if (!stripped_string.nil?()) then
        # At least one Signature removed.
        self.metadata = stripped_string
      end
    end

    # Double-check that all <Signature> elements were removed.
    if (self.metadata =~ /<([[:word:]]+:)?Signature([ ]|>)/) then
      errors.add(:base, 'Error stripping signatures; please remove any ' +
                        '<Signature> elements from your metadata before saving.')
    end

    ## ## VALIDUNTIL ATTRIBUTE
    # Check the validUntil attribute.
    # If the validUntil attribute exists and is less than our expiration limit raise a
    # validation error.
    if (not ignore_valid_until_expiration) then
      validUntil = self.valid_until
      if (not validUntil.blank?()) then
        # Lower limit
        valid_until_lower_limit_utc = now_utc.to_i() +
                                           Rails.configuration.valid_until_lower_limit_seconds
        if (validUntil.to_i() < valid_until_lower_limit_utc) then
            limit_text = Rails.configuration.valid_until_lower_limit_text
            errors.add(:base,
                       "Metadata's validUntil attribute (#{validUntil}) " \
                       "does not have at least #{limit_text} left before expiration")
        end

        # Upper limit
        valid_until_upper_limit_utc = now_utc.to_i() +
                                           Rails.configuration.valid_until_upper_limit_seconds
        if (validUntil.to_i() > valid_until_upper_limit_utc) then
            limit_text = Rails.configuration.valid_until_upper_limit_text
            errors.add(:base,
                       "Metadata's validUntil attribute is (#{validUntil}) which " \
                       "is more than the limit of #{limit_text}")
        end
      end
    end

    ## EntityID blacklist
    # Make sure the submitted entityId does not match any regular expression in
    # the entityID blacklist.
    Rails.configuration.entityid_blacklist.each do |entry_rx, entry_descriptor|
      if (self.entityid =~ /#{entry_rx}/) then
        errors.add(:base, %Q(Your entityID '#{self.entityid}' ) +
                          %Q(is reserved for '#{entry_descriptor}'))
      end
    end

    ## EntityID formatting.
    # For new records we require that the entityID be an absolute URI.
    # However, we do allow some epxcific exceptions (e.g., AWS's
    # cognito).
    #
    # See also the "InCommon Metadata Registration Practice Statement"
    # page at https://www.incommon.org/federation/mrps/
    if ((not ignore_entityid_format_issues) and self.new_record?()) then
      # This is a NEW record

      # Is the entityID valid?
      if (not Utility.entityid_valid?(self.entityid)) then
        invalid_entityid_message = <<~INVALID_ENTITYID_MESSAGE.strip.html_safe
          Your entityID &quot;#{self.entityid}&quot; is not in the form of a URI. All
          SPDB records must have an entityID that is an absolute URI starting
          with &quot;http&quot; or &quot;https&quot;.
          See also <a target="_blank"
          href="https://www.incommon.org/federation/mrps/">InCommon
          Metadata Registration Practice Statement</a>.
        INVALID_ENTITYID_MESSAGE

        errors.add(:base, invalid_entityid_message)
      else
        # This is an EXISTING record so we skip the check.
      end
    end
  end

  # Function to add a parameter (used to pass information from controller
  # to model).
  def add_groups_attributes(groups_attributes)
    log_me('entering add_groups_attributes')
    if (not groups_attributes.nil?) then
      log_me("adding groups_attributes: #{groups_attributes.inspect}")
      self.xgroups_attributes = groups_attributes
    else
      log_me('groups_attributes is nil')
      self.xgroups_attributes = {}
    end
    return
  end

  # Function to add a parameter (used to pass information from controller
  # to model).
  def add_entitlements_attributes(entitlements_attributes)
    log_me('entering add_entitlements_attributes')
    if (not entitlements_attributes.nil?) then
      log_me("adding entitlements_attributes: #{entitlements_attributes.inspect}")
      self.xentitlements_attributes = entitlements_attributes
    else
      log_me('entitlements_attributes is nil')
      self.xentitlements_attributes = {}
    end
    return
  end

  # Check that there is at least one owning group AND that none
  def at_least_one_owning_group()
    log_me('entering at_least_one_owning_group')
    # remember that xgroups_attributes
    groups_attributes = self.xgroups_attributes
    log_me("xgroups_attributes is #{xgroups_attributes.inspect}")

    # Create an array of all workgroup names that are not to be destroyed
    not_destroyed_workgroups = []
    if (groups_attributes) then
      groups_attributes.each do |key, group_attr|
        # group_attr is a hash. At least one must have its key
        # '_destroy' not be 1.
        if (group_attr.key?('_destroy')) then
          destroy_value  = group_attr['_destroy']
          workgroup_name = group_attr['workgroup']
          log_me("workgroup_name: #{workgroup_name}, destroy_value: #{destroy_value}")
          if ((destroy_value != '1') and (destroy_value != 'true')) then
            not_destroyed_workgroups.push(workgroup_name)
          end
        end
      end
    end

    # If not_destroyed_workgroups is empty, there are no owning groups.
    # That is not allowed.
    if (not_destroyed_workgroups.empty?) then
      log_me('problem: not_destroyed_workgroups has zero length')
      errors.add(:base, 'Must have at least one owning group')
      return
    else
      log_me("not_destroyed_workgroups has length #{not_destroyed_workgroups.size}")
    end

    # Make sure at least one of the workgroups is not a personal
    # workgroup.
    at_least_one_workgroup_is_not_personal = false
    not_destroyed_workgroups.each do |workgroup_name|
      if (workgroup_name !~ /^~/) then
        # This is not a personal group
        at_least_one_workgroup_is_not_personal = true
        break
      end
    end

    if (not at_least_one_workgroup_is_not_personal) then
      errors.add(:base, 'Must have at least one non-personal owning group')
    end

    # If we get here, then everything is good and we can simply return.
    return
  end

  # Check that there are not too many entitlements.
  def does_not_exceed_maximum_entitlements()
    log_me('does_not_exceed_maximum_entitlements')

    # Can skip this if allow_unlimited_entitlements is set.
    if (self.allow_unlimited_entitlements) then
      return
    else
      # remember that xentitlements_attributes
      entitlements_attributes = self.xentitlements_attributes
      log_me("xentitlements_attributes is #{xentitlements_attributes.inspect}")

      if (entitlements_attributes) then

        # Count up the number of entitlements being saved (be sure and omit
        # the ones being deleted).
        not_destroyed_entitlements = []
        entitlements_attributes.each do |key, entitlement_attr|
          # entitlement_attr is a hash. Ignore those where '_destroy' is
          # set to 1.
          if (entitlement_attr.key?('_destroy')) then
            destroy_value    = entitlement_attr['_destroy']
            entitlement_name = entitlement_attr['workgroup']
            log_me("entitlement_name: #{entitlement_name}, destroy_value: #{destroy_value}")
            if ((destroy_value != '1') and (destroy_value != 'true')) then
              not_destroyed_entitlements.push(entitlement_name)
            end
          end
        end

        max_entitlements = Rails.configuration.maximum_entitlements
        number_entitlements = not_destroyed_entitlements.length()
        if (number_entitlements > max_entitlements) then
          errors.add(:base, "Your SP has #{number_entitlements} entitlements " +
                            "which is greater than the maximum allowed of #{max_entitlements}")
        else
          return
        end
      end
    end
  end


  # Returns an a "counting" hash mapping usage type to number of
  # certificates of that type.
  #
  # The four types:
  #
  #  'signing'
  #  'encryption'
  #  'unspecified'
  #  'unrecognized'
  #
  # Example:
  #    { signing: 1, encryption: 1, unspecified: 0, unrecognized: 0 },
  #
  def get_cert_use_count(certificates)
    all_cert_use_count = {}
    all_cert_use_count['encryption']   = 0
    all_cert_use_count['signing']      = 0
    all_cert_use_count['unspecified']  = 0
    all_cert_use_count['unrecognized'] = 0

    certificates.each do |certificate|
      if (certificate.use_is_encryption?())
        usage_type = 'encryption'
      elsif (certificate.use_is_signing?())
        usage_type = 'signing'
      elsif (certificate.use_is_unspecified?())
        usage_type = 'unspecified'
      else
        usage_type = 'unrecognized'
      end

      all_cert_use_count[usage_type] += 1
    end

    return all_cert_use_count
  end

  def to_s()
    str1 = <<~TOS.strip
      entityid:           #{self.entityid}
      contact_email:      #{self.contact_email}
      owned_by:           #{self.owned_by}
      loginURL:           #{self.login_url}
      created_at:         #{self.created_at}
      updated_at:         #{self.updated_at}
      created_by:         #{self.created_by}
      updated_by:         #{self.updated_by}
      sign_responses:     #{self.sign_responses}
      sign_assertions:    #{self.sign_assertions}
      encrypt_assertions: #{self.encrypt_assertions}
    TOS
    return str1
  end

  # Write the metadata file of type metadata_type (maybe). Return true if
  # the file was written, false otherwise. Skip any configurations with
  # disabled field set to 'true'.
  def self.write_metadata_file(metadata_type, force_update = false)
    # First of all, do we even _need_ to write the metadata file?
    if (not Spconfig.should_metadata_file_be_exported(metadata_type, force_update)) then
      log_me("no need to update metadata_type '#{metadata_type}' at this time")
      return false
    else
      log_me("metadata_type '#{metadata_type}' needs to be updated")
    end

    # If we get here then we need to write the file.

    # Step 1. Define the path where the output file will go.
    filepath = self.metadata_type_to_filepath(metadata_type)

    # f will be a temporary file for the complete metadata file.
    f = Tempfile.new(metadata_type)

    # Step 2. Create a raw file of all the metadata. This will vary depending
    # on the metadata_type.
    raw_metadata_file = Tempfile.new(metadata_type)
    raw_metadata_file.close()
    Spconfig.create_raw_metadata_file(metadata_type, raw_metadata_file)

    # Step 3. Construct the complete output file.

    # Step 3a. Make the header.
    # HEADER
    # We need to replace the string '%%VALID_UNTIL%%' with a valid until
    # time in form '2016-10-10T18:43:47Z'. How long this file is valid is
    # set as a number of seconds in the application.rb configuration file.
    header = Rails.configuration.sp_metadatafile_header.dup

    expiration_time    = Time.now.utc + Rails.configuration.sp_metadatafile_expire_seconds
    valid_until_string = expiration_time.strftime('%FT%TZ')

    header.sub!('%%VALID_UNTIL%%', valid_until_string)

    # We also need to replace the string '%%NAME%%' with the correct
    # headername (see the definition of METADATA_TYPE_TO_HEADERNAME
    # above for more details).
    headername = METADATA_TYPE_TO_HEADERNAME[metadata_type]
    header.sub!('%%NAME%%', headername)

    # We also need to replace '%%ID_REPLACE%%' with a unique
    # identifier. We use a timestamp not for any technical reason but just
    # to make it easier to determine when the metadata file was created.
    # Example:
    # ID='20161014075412MDrrr'. The 'MD' is the literal string 'MD' and
    # is there to indicate this is MetaData. The 'rrr' are three random
    # letters to make extra sure there is no ID name collision.
    current_time = Time.now.utc
    rrr          = (0...3).map { ('a'..'z').to_a[rand(26)] }.join
    id_string    = current_time.strftime('%Y%m%d%H%M%S') + 'MD' + rrr

    # Now do the replace.
    header.sub!('%%ID_REPLACE%%', id_string)

    f.write(header)
    f.write("\n")

    # Step 3b. Put in all the actual metadata.
    # MIDDLE (the raw metadata strings)
    File.open(raw_metadata_file, 'r') do |rf|
      rf.each_line do |line|
        f.write(line)
      end
    end

    # Step 3c. Put in the footer.
    # FOOTER
    f.write("\n")
    f.write(Rails.configuration.sp_metadatafile_footer)

    f.close()

    log_me("wrote complete unsigned metadata file to '#{f.path()}'")

    # Step 4. Remove the raw metadata file.
    # At this point, we can delete the temporary file raw_metadata
    raw_metadata_file.unlink()

    # Step 5. Sign the output file (maybe) and move the file into place.
    if (Rails.configuration.sp_metadatafile_sign) then
      Utility.sign_xml_file(f.path(), filepath)
    else
      File.rename(f.path(), filepath)
      log_me("not signing metadata file; moved temporary unsigned metadata file to '#{filepath}'")
    end

    # Cleanup the temporary file
    if (File.exist?(f.path())) then
      f.unlink()
    end

    # We return true since the file was written.
    return true
  end

  def self.metadata_type_to_filepath(metadata_type)
    basepath = Rails.configuration.sp_metadatafile_basedir

    # Get the basename for this metadata type.
    basename = METADATA_TYPE_TO_BASENAME[metadata_type]

    basename_xml = "#{basename}.xml"
    filepath     = File.join(basepath, basename_xml)

    return filepath
  end

  # The metadata_type parameter must be one of these values:
  #    * 'combined'   - all the metadata
  #    * 'legacy'     - all the metadata whose is_legacy field is TRUE
  #    * 'non-legacy' - all the metadata whose is_legacy field is FALSE
  #
  def self.should_metadata_file_be_exported(
        metadata_type,
        force_update = false
      )
    metadata_filepath = Spconfig.metadata_type_to_filepath(metadata_type)

    return Utility.should_file_be_exported(
             filepath:           metadata_filepath,
             updatedtime_name:   'last-changed',
             force_update:       force_update,
             time_limit_seconds: Rails.configuration.sp_metadata_export_time_limit_seconds,
             action_label:       'metadata-changes'
           )
  end

  def self.create_raw_metadata_file(metadata_type, filepath)
    # Read the metadata from all the records into a file in a normalized
    # order. Returns the path to the supplied filepath. Skip any configurations
    # with disabled field set to 'true'.
    #
    # Remember that we DON'T want to include metadata from an spconfig that is
    # part of a federation.
    if (metadata_type == 'combined') then
      spconfigs = Spconfig.where(federation_id: nil,
                                 disabled: false).order(:id).all()
    elsif (metadata_type == 'legacy') then
      spconfigs = Spconfig.where(federation_id: nil,
                                 is_legacy: true, disabled: false).order(:id).all()
    elsif (metadata_type == 'non-legacy') then
      spconfigs = Spconfig.where(federation_id: nil,
                                 is_legacy: false, disabled: false).order(:id).all()
    else
      raise "unrecognized metadata type #{metadata_type}"
    end

    File.open(filepath, 'w') do |f|
      spconfigs.each do |spc|
        f.write(spc.metadata)
      end
    end

    return filepath
  end

  # Save all the metadata_types to their respective files. We return a
  # hash that maps the metadata_type to a boolean. Mapping to true
  # means the file was actually written, false means the write was
  # skipped. Note that we skip all configurations with the disabled
  # field set to 'true'.
  def self.save_metadatas_to_file(force_update: false)
    metadata_type_to_was_written = {}
    METADATA_TYPE_TO_BASENAME.each do |metadata_type, basename|
      log_me("creating (if necessary) metadata output file for metadata_type '#{metadata_type}'")
      rv = Spconfig.write_metadata_file(metadata_type, force_update)
      metadata_type_to_was_written[metadata_type] = rv
    end
    return metadata_type_to_was_written
  end

  # Return true if user is in at least one of this object's groups
  def old_user_in_groups(user)
    groups.each do |group|
      log_me("checking to see if #{user} is in group #{group.workgroup}")
      privgroup = Privgroup.new(group.workgroup)
      if (privgroup.user_is_member?(user))
        return true
      end
    end

    return false
  end

# Mimicing old user_in_groups by iterating through SP's owning workgroups
# And intersecting with user's owning workgroup
  def user_in_groups(user,my_owning_workgroups)
    sp_groups = []
    status = false 
    if (! my_owning_workgroups.empty?) then
      groups.each do |group|
         sp_groups.push(group.workgroup)
      end
      intersect_group =  sp_groups & my_owning_workgroups
      log_me("checking to see if #{user} is in sp #{sp_groups} and my_owning #{my_owning_workgroups}")
      if ( ! intersect_group.empty?) then
        status = true
      else
        status = false 
      end
    end
    return status
  end

  # Return an array of the spconfigs that the given user has access
  # to. That is all the spconfigs that the user owns, or that contain a
  # group that the user is a member of.
  def self.this_users_spconfigs(user)
    # Step 1. Get all the spcongifs owned by this user.
    spconfigs_owned = []
    spconfigs_owned = Spconfig.where(owned_by: user)

    # 2. Now we find all spconfigs that have an associated group
    # containing user's owning_workgroup
    my_owning_workgroups = []
    my_owning_workgroups = Utility.get_my_owning_workgroups(user)
    #log_me("#{my_owning_workgroups}")

    spconfigs = []
    if (! my_owning_workgroups.empty?) then
      #users = User.includes(:address)
      spconfigs_all = Spconfig.includes(:groups).all()
      spconfigs_group_owned = []
      
      spconfigs_all.each do |spconfig|
        groups1 = spconfig.groups
        sp_groups = []
        groups1.each do |group|
          sp_groups.push(group.workgroup)
        end
  
        if (!sp_groups.empty?) then
          intersect_group = sp_groups & my_owning_workgroups
          if (!intersect_group.empty?)
            spconfigs_group_owned.push(spconfig)
          end
        end

      end #### ending loop for spconfigs_all
  
      spconfigs = spconfigs_owned + spconfigs_group_owned
    end

    if (!spconfigs.empty?) then
      spconfigs = spconfigs.uniq()
    end
    return spconfigs

  end

  # Return true of the current logged-in user has "control" of this
  # spconfig, that is, is allowed to update or delete the spconfig.
  #
  # A user has control if he satisfies one of the following:
  #   1. The user is an SPDB admin
  #   2. The owner of this spconfig is the user.
  #   3. The user is a member of one of the groups that owns this spconfig.
  def user_controls?(user)
    my_owning_workgroups = []
    my_owning_workgroups = Utility.get_my_owning_workgroups(user)
    if (Utility.admin?(user)) then
      return true
    elsif (user == self.owned_by) then
      return true
    elsif (self.user_in_groups(user,my_owning_workgroups)) then
      return true
    else
      return false
    end
  end

  def contact_is_stanford_address?()
    if (self.contact_email.nil?()) then
      return false
    end

    address = self.contact_email
    if (address =~ /^.*([.]|@)stanford[.]edu$/i) then
      return true
    else
      return false
    end
  end

  def contact_is_stanford_list_address?()
    if (self.contact_email.nil?()) then
      return false
    end

    address = self.contact_email
    if (address =~ /^.*@lists[.]stanford[.]edu$/i) then
      return true
    else
      return false
    end
  end

  # Return an array of the workgroup names. Will return the empty
  # array if there are none.
  def workgroup_names()
    if (groups.nil?() or groups.empty?())
      return []
    else
      return groups.map { |g| g.workgroup }
    end
  end

  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###
  ### entitlements file writing section
  def self.save_entitlements_attribute_filter_file(force_update = false)
    # Do we even _need_ to write the entitlements file?
    if (not Entitlement.should_entitlements_file_be_exported(force_update)) then
      log_me('no need to update entitlement file at this time')
      return false
    else
      log_me('entitlement file needs to be updated')
    end
    sleep 30
    Spconfig.write_entitlements_attribute_filter_file(Entitlement.entitlements_file_filepath())
  end

  # Outputs a (possibly) signed auto-entitlements file at the location filepath.
  # On error raises an exception.
  def self.write_entitlements_attribute_filter_file(filepath)
    # Step 1. Create the unsigned entitlements file.
    f = Tempfile.new('entitlementsxml')
    f.write(Spconfig.entitlement_attribute_filter_string())
    f.close()

    # Step 2. Verify that the file created in Step 1 has valid XML.
    if (!Utility.valid_xml?(f.path())) then
      if (File.exist?(f.path())) then
        f.unlink()
      end
      raise "the entitlements file is not valid XML"
    end

    # Step 3. If configured to do so, sign the file created in Step 1.
    # Construct path for output file
    basedir     = Rails.configuration.sp_entitlements_xml_file_basedir
    basename    = Rails.configuration.sp_entitlements_xml_file_basename
    output_path = File.join(basedir, basename)

    if (Rails.configuration.sp_entitlements_xml_file_sign) then
      Utility.sign_xml_file(f.path(), output_path)
    else
      # Since not signing simply copy temp file to output_path.
      FileUtils.cp(f.path(), output_path)
    end

    if (File.exist?(f.path())) then
      f.unlink()
    end
  end

  def self.entitlement_attribute_filter_string()
    # Create the entitlement attribute file. Returns the path to the
    # supplied filepath.

    id_string = 'SPDB_Auth_Entitlement_Release'

    entitlement_xml_header =
      <<~HEREDOC
      <?xml version="1.0" encoding="UTF-8"?>
      <AttributeFilterPolicyGroup id="#{id_string}"
                                  xmlns="urn:mace:shibboleth:2.0:afp"
                                  xmlns:saml="urn:mace:shibboleth:2.0:afp:mf:saml"
                                  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                                  xsi:schemaLocation="urn:mace:shibboleth:2.0:afp classpath:/schema/shibboleth-2.0-afp.xsd
                                  urn:mace:shibboleth:2.0:afp:mf:basic classpath:/schema/shibboleth-2.0-afp-mf-basic.xsd
                                  urn:mace:shibboleth:2.0:afp:mf:saml classpath:/schema/shibboleth-2.0-afp-mf-saml.xsd">
      HEREDOC

    entitlement_xml_header_extra =
      <<~HEREDOC
      <!--
        PLEASE READ:

        This file was automatically generated by the SPDB application
        and will be overwritten the next time the SPDB exports this
        information.

        This file contains "eduPersonEntitlement" attribute release
        rules for those SPs in the SPDB that have "entitlements" added
        to their records. Note that an SP appearing in this file will
        also appear in another attribute release file if it has
        attributes beyond eduPersonEntitlement that need to be
        released.

        Each entry's id is of the form id="su_edupe_MD5SUM" where
        "MD5SUM" is the MD5 sum of the entry's entityID.

        This file was generated on #{Time.now}
      -->
      HEREDOC

    entitlement_xml_footer =
      <<~HEREDOC
      </AttributeFilterPolicyGroup>
      HEREDOC

    entitlement_string = ''
    entitlement_string += entitlement_xml_header + "\n"
    entitlement_string += entitlement_xml_header_extra.indent(4) + "\n"
    entitlement_string += Spconfig.create_entitlement_attribute_filter_string_raw.indent(4)
    entitlement_string += entitlement_xml_footer

    return entitlement_string
  end

 # Create the raw entitlements filter. Skip any configuration with
 # disabled field set to 'true' or allow_entitlement_release set to
 # 'false'.
 def self.create_entitlement_attribute_filter_string_raw()
    # Sort the spconfigs by entityid
    spconfigs = Spconfig.joins(:entitlements).distinct.order(entityid: :asc)

    result = ''
    spconfigs.each do |spc|
      if (!spc.disabled && spc.allow_entitlement_release) then
        result += spc.entitlement_to_xml
      end
    end

    return result
  end

  # Return an array of the entitlement names. Will return the empty
  # array if there are none.
  def entitlement_names()
    if (entitlements.nil?() or entitlements.empty?())
      return []
    else
      return entitlements.map { |g| g.workgroup }
    end
  end
  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###  ###

  def make_help_request_url(extra = 'SPDB')
    return Utility.make_help_request_url(self.entityid, extra)
  end

  def after_create_callback
    log_me('configuration created; touching the updatedtimes table')
    Updatedtime.touch()
  end

  def after_destroy_callback
    log_me('configuration deleted; touching the updatedtimes table')
    Updatedtime.touch()
  end

  # Call this method AFTER any save.
  def after_save_callback
    change_detected = false

    # Did the 'metadata' field change?
    if (self.saved_change_to_metadata?) then
      change_detected = true
      log_me(%q(change in 'metadata' field detected))
    end

    # Did the 'is_legacy' field change?
    if (self.saved_change_to_is_legacy?) then
      change_detected = true
      log_me(%q(change in 'is_legacy' field detected))
    end

    # Did the 'disabled' field change?
    if (self.saved_change_to_disabled?) then
      change_detected = true
      log_me(%q(change in 'disabled' field detected))
    end

    if (change_detected) then
      log_me('XML-affecting change detected; touching the updatedtimes table')
      Updatedtime.touch()
    else
      log_me('No XML-affecting change detected; not touching the updatedtimes table')
    end

    # Remove any duplicate workgroups.
    workgroup_hash = {}
    self.groups.each do |group|
      workgroup = group.workgroup
      if (workgroup_hash.key?(workgroup)) then
        # This is already in the list, so delete
        group.delete()
        log_me("deleting duplicate workgroup '#{workgroup}'")
      else
        workgroup_hash[workgroup] = true
      end
    end

    # Remove any duplicate entitlements.
    entitlement_hash = {}
    self.entitlements.each do |entitlement|
      workgroup = entitlement.workgroup
      if (entitlement_hash.key?(workgroup)) then
        # This is already in the list, so delete
        entitlement.delete()
        log_me("deleting duplicate entitlement '#{workgroup}'")
      else
        entitlement_hash[workgroup] = true
      end
    end

    # Add the certificates (if any)
    certificates = Utility.get_certificate_objects_from_metadata(metadata)

    self.certificates.destroy_all
    self.certificates << certificates
  end

  # Call this method BEFORE any save.
  def before_save_callback
    # Only need to call this method if there is no federation (remember:
    # beign part of a federation means no metadata).
    if (self.federation) then
      return
    end

    # Remove the XML declaration <?xml ...> if there is one.
    self.metadata = try_to_strip_xml_declaration(self.metadata)
  end

  def entityid_md5sum()
    return Digest::MD5.hexdigest(self.entityid)
  end

  # Generate the string used in the entitlement attribute filter file for this
  # SP. Returns nil if this SP has not entitlements to release.
  def entitlement_to_xml()
    # Create a unique XML id
    xml_id = "su_edupe_#{self.entityid_md5sum()}"

    spdb_url = "https://#{Utility.my_domain_name()}"

    # Get all of this spconfig's entitlements. Sort them by workgroup name.
    entitlements = self.entitlements.sort { |a, b| a.workgroup <=> b.workgroup }

    # Create the PermitValueRule clause from the entitlement values.
    rules = []
    spacer = ' ' * (3 * 4)
    entitlements.each do |entitlement|
      cur_rule = spacer + %Q(<Rule xsi:type="Value" value="#{entitlement.workgroup}" />)
      rules.push(cur_rule)
    end

    # "Sanitize" entityid so that it does not contain "--" (since "--" cannot
    # appear inside an XML comment).
    entityid_sanitized = entityid.gsub(/--/, 'DASHDASH')

    all_rules = rules.join("\n")
    <<~HEREDOC
    <!--
      From SP with entityID "#{entityid_sanitized}".
      See also #{spdb_url}/spconfigs/#{self.id}
      -->
    <AttributeFilterPolicy id="#{xml_id}">
        <PolicyRequirementRule xsi:type="Requester" value="#{entityid}" />
        <AttributeRule attributeID="eduPersonEntitlement">
            <PermitValueRule xsi:type="OR">
    #{all_rules}
            </PermitValueRule>
        </AttributeRule>
    </AttributeFilterPolicy>

    HEREDOC
  end

  def incommon_metadata_url
    return Utility.incommon_metadata_url(entityid)
  end

  protected

  def get_validation_results
    # No validation needed if spconfig is a federation stub entry.
    if (self.federation) then
      log_me("entityid is #{self.entityid}")
      return
    end

    metadata_stripped = try_to_strip_xml_declaration(metadata)

    @metadata_validation_results = Utility.submit_metadata_for_validation(metadata_stripped)

    @sp_info                = @metadata_validation_results[0]
    @saml_validation_errors = @metadata_validation_results[1]

    log_me("metadata validation results: #{@metadata_validation_results.inspect}")
    # Pre-populate entityid so we we can find it.
    if (@sp_info.key?('entityID'))
      self.entityid = @sp_info['entityID']
    end

    # Also populate validUntil (if it exists)
    if (@sp_info.key?('validUntil') and !@sp_info['validUntil'].empty?()) then
      begin
        self.valid_until = DateTime.xmlschema(@sp_info['validUntil'])
      rescue
        # We don't need to add anything to errors.add since the XML
        # schema validation will do that for us. Instead, we just trap
        # and swallow the error.
      end
    else
      self.valid_until = nil
    end
  end

  # Returns the XML without the <?xml header if the XML is valid. If the
  # XML is _not_ valid simply returns the original XML string.
  def try_to_strip_xml_declaration(metadata_string)
    stripped_metadata = nil
    begin
      stripped_metadata = Utility.strip_xml_declaration(metadata_string)
    rescue StandardError
      stripped_metadata = nil
    end

    if (stripped_metadata.nil?()) then
      return metadata_string
    else
      return stripped_metadata
    end
  end
end
