# Load all the workgroup/WA role mappings from a specified file.
# Run like this:
#
#    bin/rails runner lib/load-awsmappings.rb \
#      <path_to_amazon_attributes_file> (skip|noskip)
#
# Examples:
#
#    bin/rails runner lib/load-awsmappings.rb \
#      /tmp/attribute-resolver-amazon-auto.xml noskip
#    bin/rails runner lib/load-awsmappings.rb \
#      /tmp/attribute-resolver-amazon-auto.xml skip
#
# skip:   means to NOT validate workgroups against the configured ldap server.
# noskip: means to validate workgroups against the configured ldap server.
#
# To copy all the metadata fragment files:
#
# scp -r3 root@idp08:/etc/shibboleth-idp/conf/attribute-resolver-amazon-auto.xml \
#   root@lemon-uat:/srv/scratch/cheroux/

require 'utility'

# Stage 1. Get the awsaccount mapping info from the file.

if (ARGV[0]) then
  xmlfile = ARGV[0]
else
  raise 'missing aws mappings XML file'
end

skip_workgroup_validation = true
if (ARGV[1]) then
  skip = ARGV[1]

  if (skip == 'skip') then
    skip_workgroup_validation = true
  elsif (skip == 'noskip') then
    skip_workgroup_validation = false
  else
    raise "second argument should be either 'skip' or 'noskip'"
  end
else
  raise 'missing skip/noskip argument'
end

xmlfile_string = File.open(xmlfile, 'r').read

# Parse the file
xml_doc = Nokogiri::XML(xmlfile_string).remove_namespaces!()

# Get all ValueMaps
valuemaps = xml_doc.xpath('//ValueMap')

maps = []
awsaccount_numbers = {}

valuemaps.each do |valuemap|
  returnvalue = valuemap.xpath('./ReturnValue').text()
  sourcevalue = valuemap.xpath('./SourceValue').text()

  provider, role = returnvalue.split(',')
  workgroup      = sourcevalue

  puts "provider is  #{provider}"
  puts "role is      #{role}"
  puts "workgroup is #{workgroup}"

  # Get the account number. Should be the same in bothj provider and role.
  # arn:aws:iam::369030361594:saml-provider/stanford-idp
  if (provider =~ %r{::(\d{12}):}) then
    number = $1
    puts "number is #{number}"
  else
    raise 'could not parse provider'
  end

  # arn:aws:iam::369030361594:role/AdministratorAccess-generatedby-stanford-sso
  if (role =~ %r{::(\d{12}):}) then
    role_number = $1
    if (role_number != number) then
      raise 'account number is role does not match that in provider'
    end
  else
    raise 'could not parse provider'
  end

  instance = Awsaccount.arn_to_instance(provider)
  awsaccount_numbers[number] = instance

  # We only store the "short" name of the role, i.e., the last part.
  if (role =~ %r{\/(.*)$}) then
    short_role = $1
  else
    raise "could not find role name in #{role}"
  end

  # collect the data into an array
  maps.push([number, provider, short_role, workgroup])
end

# Stage 2. Delete all the awsaccounts and add them back.

rv = Awsaccount.delete_all()
puts "number of records deleted #{rv}"

acct_objects = []
default_owning_workgroup = 'acs:team'
$remote_user = 'auto-migration'
awsaccount_numbers.each do |number, instance|
  puts "about to add [#{number}, #{instance}]"

  awsacct = Awsaccount.new()
  awsacct.instance = instance
  awsacct.number   = number
  awsacct.nickname = 'auto-migrated'

  awsacct.save()

  aws_owning_group = Awsowninggroup.new()
  aws_owning_group.skip_workgroup_validation = skip_workgroup_validation

  aws_owning_group.workgroup = default_owning_workgroup
  awsacct.awsowninggroups << aws_owning_group

  awsacct.save()

  acct_objects.push(awsacct)
end

acct_objects.each do |awsacct|
  puts awsacct.to_s()
end

# Stage 3. Add the mappings
maps.each do |map|
  number    = map[0]
  provider  = map[1]
  role      = map[2]
  workgroup = map[3]

  # Get the account with the given number
  acct = Awsaccount.find_by_number(number)
  if (acct.nil?()) then
    raise "could not find awsaccount with number #{number}"
  end

  # Verify that the provider matches
  providers = acct.awsidproviders
  if (providers.length() > 1) then
    raise "account #{number} has too many id providers"
  end

  acct_provider = providers[0]

  if (acct_provider.arn != provider) then
    raise "account #{number} has provider #{providers[0]} that does not match #{provider}"
  end

  # Add the mapping.
  puts "creating rolemapping (#{workgroup}, #{role})"

  arm = Awsrolemapping.new()
  arm.skip_workgroup_validation = skip_workgroup_validation

  arm.workgroup = workgroup
  arm.role      = role

  acct_provider.awsrolemappings << arm

  acct_provider.save()
  acct.save()
  puts "saved (#{role}, #{workgroup})"
  puts acct.to_s()
end

Awsaccount.all().each do |awsacct|
  puts awsacct.to_s()
end
