require 'utility'
require 'logger'
require 'tempfile'

class Privgroup

  @@LDAP_COMPARE_FALSE = 5
  @@LDAP_COMPARE_TRUE  = 6

  def initialize(name)
    @name = name

    @use_caching = Rails.configuration.ldap_cache_enable

    @cache_duration = Rails.configuration.ldap_cache_duration
    if (@cache_duration.nil?()) then
      @cache_duration = "10.minutes"
    end

    @private_privgroup_set = Privgroup.read_private_from_file()

    @private = @private_privgroup_set.include?(@name)
  end

  # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###

  # Returns true if user is a member of privgroup, false otherwise.
  #
  # The third parameter, administrator_flag, if set to true will verify
  # that the user is an administrator of the privgroup rather than a
  # member of the privgroup. Remember that a user can be just a member,
  # just an administrator, or both (or, of course, neither).  This flag is,
  # by default, false.
  #
  # def self.user_member_of_privgroup?(user, privgroup, administrator_flag = false)
  def user_is_member?(user, administrator_flag = false)
    Utility.logger.debug("checking to see if #{user} is a member of #{@name}")
    if (@use_caching) then
      Utility.logger.debug("will check using caching")
      return self.user_is_member_cached?(user, administrator_flag)
    else
      Utility.logger.debug("will check without using caching")
      return self.user_is_member_noncached?(user, administrator_flag)
    end
  end

  def user_is_member_noncached?(user, administrator_flag = false)
    Utility.logger.debug("doing non-cached check if {@user} is a member of {@name}")
    ldap_server     = Rails.configuration.ldap_service_name
    user_DN         = "uid=#{user},cn=accounts,dc=stanford,dc=edu"
    ldapcompare_cmd = '/usr/bin/ldapcompare'

    if (administrator_flag) then
      match_string = "suPrivilegeGroup:#{@name}/administrator"
    else
      match_string = "suPrivilegeGroup:#{@name}"
    end

    # The command looks like this:
    #
    #   /usr/bin/ldapcompare -Q -z -h ldap.stanford.edu \
    #   uid=adamhl,cn=accounts,dc=stanford,dc=edu suPrivilegeGroup:acs:onedrive-share
    #
    # or, in the case when administrator_flag is true,
    #
    #   /usr/bin/ldapcompare -Q -z -h ldap.stanford.edu \
    #   uid=adamhl,cn=accounts,dc=stanford,dc=edu suPrivilegeGroup:acs:onedrive-share/administrator
    #
    cmd = %Q('#{ldapcompare_cmd} -Q -z -h #{ldap_server} "#{user_DN}" #{match_string}')

    (stdout, stderr, rc) = Utility.run_cmd_under_service_principal(cmd)
    Utility.logger.debug("(#{stdout}, #{stderr}, #{rc})")

    if (rc == @@LDAP_COMPARE_TRUE) then
      if (administrator_flag) then
        Utility.logger.debug("user #{user} IS an administrator of group #{@name}")
      else
        Utility.logger.debug("user #{user} IS in group #{@name}")
      end
      return true
    elsif (rc == @@LDAP_COMPARE_FALSE) then
      if (administrator_flag) then
        Utility.logger.debug("user #{user} is NOT an administrator of group #{@name}")
      else
        Utility.logger.debug("user #{user} is NOT in group #{@name}")
      end
      return false
    else
      # Some other error. What should we do?
      return false
    end

  end

  def cache_key(user, administrator_flag = false)
     if (administrator_flag) then
       return "#{user}::#{@name}::administrator"
     else
       return "#{user}::#{@name}::member"
     end
  end

  def user_is_member_cached?(user, administrator_flag = false)
    cache_key  = self.cache_key(user, administrator_flag)
    cache_miss = false
    expiration = @cache_duration

    # Remember that passing a block means that if fetch results in a cache
    # miss then the last value of the block will be returned.
    value = Rails.cache.fetch(cache_key, expires_in: expiration) do
      cache_miss = true
      self.user_is_member_noncached?(user, administrator_flag)
    end

    if (cache_miss) then
      Utility.logger.info("cache miss for #{cache_key}")
    else
      Utility.logger.debug("cache hit for #{cache_key}")
    end

    return value
  end
  # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###

  # Returns true if privgroup is a valid privgroup, that is, if privgroup
  # is in the cn=groups,dc=stanford,dc=edu tree.
  def valid?()
    # We keep a separate list of system-maintained groups.
    if (self.system_maintained?()) then
      return true
    end

    ldap_server    = Rails.configuration.ldap_service_name
    group_DN       = "cn=#{@name},cn=groups,dc=stanford,dc=edu"
    ldapsearch_cmd = '/usr/bin/ldapsearch'

    cmd = %Q('#{ldapsearch_cmd} -Q -LLL -h #{ldap_server} -b #{group_DN} entryDN')
    (stdout, stderr, rc) = Utility.run_cmd_under_service_principal(cmd)

    if (rc == 0) then
      Utility.logger.debug("privgroup #{@name} found")
      return true
    else
      # Some other error. What should we do?
      Utility.logger.debug("privgroup #{@name} not found")
      return false
    end

  end

  # Due to the fact some privgroups are not in the groups tree we read
  # a list of these System Maintained groups from a text file. The current
  # list of such System Maintained groups is at this web page:
  # https://asconfluence.stanford.edu/confluence/pages/viewpage.action?pageId=1168015483
  def system_maintained?()
    system_maintained_groups_path = "/etc/spdb/system-maintained-groups"
    if (File.exist?(system_maintained_groups_path)) then
      groups_raw = File.read("/etc/spdb/system-maintained-groups")

      # Split into lines
      groups = groups_raw.split("\n")

      return groups.include?(@name)
    else
      return false
    end
  end


  # Return true if the user is either a member of, or an administrator of,
  # the Stanford Workgroup privgroup; returns false otherwise.
  def user_has_access?(user)
    is_member        = self.user_is_member?(user, false)
    is_administrator = self.user_is_member?(user, true)

    return (is_member || is_administrator)
  end


  # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###

  def private?()
    Utility.logger.debug("starting private?()")

    # We consider system-maintained groups to be non-private (this was
    # confirmed by MAIS (see also
    # https://asjira.stanford.edu/jira/browse/ISHIB-633).
    if (self.system_maintained?()) then
      Utility.logger.info("privgroup #{@name} is a system maintained privgroup and as such is not private")
      return false
    else
      return @private
    end
  end

  def self.read_private_from_file()
    Utility.logger.debug("read private privgroup info from file")

    destination = Rails.configuration.private_privgroups_path

    if (! File.file?(destination)) then
      # If the file does not exist fall back to getting the information
      # directly from the LDAP.
      Utility.logger.warn("no file '#{destination}' so will have to read LDAP directly")
      return Privgroup.get_all_private_from_ldap()
    else
      privgroups_raw = File.read(destination)

      # Split into lines
      privgroups = privgroups_raw.split("\n")

      private_privgroups_set = Set.new()
      privgroups.each do |privgroup|
        private_privgroups_set.add(privgroup)
      end

      Utility.logger.debug("finished reading private privgroup info from file")

      return private_privgroups_set
    end
  end


  def self.get_all_private_from_ldap()
    Utility.logger.debug("entering get_all_private_from_ldap")

    ldap_server    = Rails.configuration.ldap_service_name
    group_DN       = "cn=groups,dc=stanford,dc=edu"
    ldapsearch_cmd = '/usr/bin/ldapsearch'
    filter         = "suVisibWorkgroup=private"

    cmd = %Q('#{ldapsearch_cmd} -o ldif-wrap=no -Q -LLL -h #{ldap_server} -b #{group_DN} #{filter} entryDN')
    (stdout, stderr, rc) = Utility.run_cmd_under_service_principal(cmd)

    if (rc != 0) then
      raise "problem running ldapsearch; contact a sysadmin"
    end

    private_privgroups_set = Set.new()

    lines = stdout.split("\n")
    lines.each do |line|
      if (line =~ /entryDN: cn=(?<privgroup>\S+),cn=groups,dc=stanford,dc=edu/)
        privgroup = $1
        Utility.logger.debug("found private privgroup '#{privgroup}'")
        private_privgroups_set.add(privgroup)
      end
    end

    return private_privgroups_set
  end

  def self.write_private_privgroups()
    destination = Rails.configuration.private_privgroups_path

    private_privgroups_set = Privgroup.get_all_private_from_ldap()

    num_found = private_privgroups_set.size()
    if (num_found < 3) then
      raise "only #{num_found} private privgroups found; this is not enough; aborting"
    end

    tempfile = Tempfile.new('privgroup')
    tf_path  = tempfile.path
    Utility.logger.debug("created temporary file '#{tf_path}'")

    File.open(tempfile, 'w') do |tf|
      Utility.logger.debug("opening temporary file '#{tf_path}'")

      private_privgroups_set.each do |privgroup_name|
        tf.puts privgroup_name
      end
      Utility.logger.debug("finished writing to temporary file '#{tf_path}'")

      tf.close
      Utility.logger.debug("closed temporary file '#{tf_path}'")

      # Rename the file.
      File.rename(tf_path, destination)
      Utility.logger.debug("renamed temporary file '#{tf_path}' to '#{destination}'")
    end

    # Check that tempfile was properly renamed.
    if (File.file?(tf_path)) then
      raise "#{tf_path} was supposed to be renamed!"
    end

    return destination
  end

  # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###

end
