namespace :notifications do

  desc "task to generate e-mail notifications"
  task :mails => :environment do

    emails  = {}

    # Get all the spconfigs grouped by contact_email
    query = 'SELECT entityid, contact_email FROM spconfigs'
    results = ActiveRecord::Base.connection.execute(query)
    results.each do |result|
      entityid      = result[0]
      contact_email = result[1]
      puts "contact_email is [#{contact_email}]"

      # If there is no contact_email, skip
      if (contact_email.nil? || contact_email.empty?) then
        puts "no contact_email so skipping #{entityid}"
        next
      else
        if (emails.key?(contact_email)) then
          emails[contact_email].push(entityid)
        else
          emails[contact_email] = [entityid]
        end
      end
    end

    emails.each do |contact_email, entityids|
      result = NotificationMailer.testing_notification_email(contact_email, entityids, false).deliver()
      puts "result: #{result.inspect()}"
    end
  end

  ######################################################################
  #
  # Generate an e-mail if a certificate is set to expire.
  #
  # From command line, to run with day_window set to 30 days:
  #
  #    bin/rake notifications:cert_expirations[30] RAILS_ENV=development
  #
  # or
  #
  #    bin/rake notifications:cert_expirations[30]
  #
  # To override e-mail:
  #
  #    bin/rake notifications:cert_expirations[30,adamhl@stanford.edu]
  #
  # To specify the environment:
  #
  #    RAILS_ENV=production bin/rake notifications:cert_expirations[30,adamhl@stanford.edu]

  desc "task to generate e-mail notifications of certificate expirations"
  task :cert_expirations, [:day_window,:email_override] => :environment do |t, args|

    day_window = args[:day_window]
    if (day_window.nil?()) then
      raise "missing argument for day_window"
    else
      day_window = args['day_window'].to_i
    end

    email_override = args[:email_override]

    now_utc            = Time.now.utc
    warning_date       = now_utc + day_window.days
    warning_date_plus1 = now_utc + (day_window + 1).days

    spconfigs = Spconfig.all()

    spconfigs.each do |spconfig|
      certificates = spconfig.certificates

      found_expiring_certificate = false
      certificates.each do |certificate|
        # Get expiration date
        if ((certificate.expires_at < warning_date_plus1) and
            (certificate.expires_at >= warning_date)) then
          found_expiring_certificate = true
        end
      end
      if (found_expiring_certificate) then
        # puts "one or more certificates associated with #{spconfig.entityid} are expiring in #{day_window} days"
        result = CertExpirationMailer.warning_email(spconfig:       spconfig,
                                                    day_window:     day_window,
                                                    email_override: email_override).deliver()
        # puts "result: #{result.inspect()}"
      end
    end

  end

  ######################################################################
  #
  # Generate an e-mail if an SP configuration is set to expire (based on
  # validUntil).
  #
  # From command line, to run with day_window set to 30 days:
  #
  #    bin/rake notifications:spconfig_expirations[30] RAILS_ENV=development
  #
  # or
  #
  #    bin/rake notifications:spconfig_expirations[30]
  #
  # To override e-mail:
  #
  #    bin/rake notifications:spconfig_expirations[30,adamhl@stanford.edu]
  #
  # To specify the environment:
  #
  #    RAILS_ENV=production bin/rake notifications:spconfig_expirations[30,adamhl@stanford.edu]
  #
  # If day_window is set to nil then rather than look for expirations
  # that fall within a certain windw we report on all expired
  # configurations.
  #
  # We skip sending notifications for configurations that are marked disabled.

  desc "task to generate e-mail notifications of SP configuration expirations"
  task :spconfig_expirations, [:mode,:dry_run,:day_window,:email_override] => :environment do |t, args|

    debug = false
    dbg = ->(msg) { if (true) then print("#{msg}\n") end }

    mode           = args[:mode]
    dry_run        = args[:dry_run]
    day_window     = args[:day_window]
    email_override = args[:email_override]

    valid_modes = ['expired', 'warning']
    if (not valid_modes.include?(mode)) then
      raise "error: mode must be one of " +  valid_modes.join(' or ')
    end

    # If dry run is not defined, raise an error.
    if (dry_run.nil?()) then
      raise "error: the dry_run argument must be supplied"
    elsif (dry_run =~ /^f/i)
      dry_run = false
    else
      dry_run = true
    end

    if (dry_run) then
      dbg.call("dry_run is true")
    else
      dbg.call("dry_run is false")
    end


    if (mode == 'expired')
      only_expired = true
    else
      only_expired = false
    end

    dbg.call("mode is #{mode}")

    now_utc = Time.now.utc

    if (not only_expired) then
      if (day_window.nil?()) then
        raise "error: missing day_window argment"
      end

      dbg.call("setting warning data window")
      day_window = day_window.to_i()
      warning_date       = now_utc + day_window.days
      warning_date_plus1 = now_utc + (day_window + 1).days
      dbg.call("warning data window is (#{warning_date}, #{warning_date_plus1}")
    end

    # Get all spconfigs that are not disabled.
    spconfigs = Spconfig.where(disabled: false).order(:id).all()

    spconfigs.each do |spconfig|
      valid_until = spconfig.valid_until

      # Skip sp configurations with blank validUntil's.
      if (valid_until.blank?()) then
        next
      end

      if (only_expired) then
        # EXPIRED CONFIGURATIONS
        if (valid_until < now_utc) then
          send_notification = true
        else
          send_notification = false
        end
      else
        # IN DATE WINDOW CONFIGURATIONS
        if ((valid_until < warning_date_plus1) and
            (valid_until >= warning_date)) then
          send_notification = true
        else
          send_notification = false
        end
      end

      if (send_notification) then
        dbg.call("sending notification for #{spconfig.id}")
        result = SpconfigExpirationMailer.warning_email(spconfig:       spconfig,
                                                        only_expired:   only_expired,
                                                        day_window:     day_window,
                                                        email_override: email_override)
        if (dry_run) then
          print result.to_s()
        else
          print result.deliver()
        end
      end
    end
  end


  # Generate list of configs along with their expirations
  desc "task to generate list of configs with their expirations"
  task :list_expirations do
    certificates = Certificate.all().order(:expires_at)
    certificates.each do |certificate|
      spconfig = certificate.spconfig
      use = certificate.use
      puts spconfig.entityid
      puts "  #{certificate.expires_at.to_s()} (#{use})"
    end
  end


end
