# frozen_string_literal: true

require 'fileutils'
require 'open3'
require 'logging'
require 'open-uri'
require 'nokogiri'
require 'net/http'
require 'net/https'
require 'uri'
require 'json'
require 'erb'

class Utility
  extend Logging

  # Logger output can be found in Apache's error.log.

  # Parse an SP metadata string and extact contact information.
  #
  #
  # <md:ContactPerson contactType="technical">
  #   <md:GivenName>Support</md:GivenName>
  #   <md:EmailAddress>winsystems@lists.stanford.edu</md:EmailAddress>
  # </md:ContactPerson>
  #
  #
  #
  # Returns a hash with keys 'GivenName' and 'EmailAddress'.
  #
  # Note Well: If there is more than one ContactPerson element, only looks
  # at the FIRST one
  def self.get_contact_info(metadata)
    result = {}

    # Get the first ContactPerson element (if there is one at all)>
    xml_doc = Nokogiri::XML(metadata).remove_namespaces!()

    nodes = xml_doc.xpath('//ContactPerson')

    if (nodes.length.positive?) then
      node = nodes.first

      contact_person_doc = Nokogiri::XML(node.to_s())

      # Get GivenName and EmailAddress
      given_name = contact_person_doc.xpath('//GivenName')
      if (given_name) then
        result['GivenName'] = given_name.text()
      end

      email_address = contact_person_doc.xpath('//EmailAddress')
      if (email_address) then
        result['EmailAddress'] = email_address.text()
      end
    end

    return result
  end

  # Strip any signatures from the metadata.
  #
  # Returns: Case 1. If there are any signatures, it will return the
  # submitted metadata with all signatures stripped. If the metadata is
  # not valid XML will raise an exception.
  #
  # Case 2. This function returns nil in two cases.
  #   Case 2a. There are no signatures in the submitted metadata.
  #   Because Nokogiri will reformat the XML string, if no signature nodes
  #   are found we return nil so that the caller does not have to re-save a
  #   version of the string that is the same except for indenting
  #   differences.
  #
  #   Case 2b. The Rails application has disabled signature stripping.

  def self.strip_signatures(metadata)
    # If we have configured our application NOT to strip signatures, then
    # simply return nil.
    if (not Rails.configuration.sp_strip_signatures) then
      Utility.logger.info('stripping signatures globally disabled; returning nil')
      return nil
    end

    # Create the XML doc object from the metadata string.
    # We set the "noblanks" option so that we don't end up with
    # unsightly blanks lines after deleting signature nodes.
    xml_doc = Nokogiri::XML(metadata) do |config|
      config.strict.noblanks
    end

    # Find the namespace prefix and URL. It will be the one with a URL
    # containing the string 'xmldsig'.
    namespaces = xml_doc.collect_namespaces

    ns_prefix = nil
    ns_url    = nil
    namespaces.each do |namespace, url|
      if (url =~ /xmldsig/) then
        matches = /xmlns:(.*)$/.match(namespace)
        if (matches) then
          ns_prefix = matches[1]
          ns_url    = url
        end
      end
    end

    if (ns_prefix.nil?()) then
      # Use an xpath search string that grabs ALL elements
      # with the name "Signature" regardless of namespace.
      xpath_string = "//*[local-name()='Signature']"
    else
      xpath_string = "//#{ns_prefix}:Signature"
    end

    # Remove any Signature nodes, keeping track of how many we remove.
    counter = 0
    if (ns_prefix.nil?()) then
      # No namespace prefix
      signature_nodes = xml_doc.xpath(xpath_string)
    else
      # There is a namespace prefix
      signature_nodes = xml_doc.xpath(xpath_string, ns_prefix => ns_url)
    end

    signature_nodes.each do |signature_node|
      signature_node.remove()
      counter += 1
    end

    if (counter.zero?) then
      # No signature nodes removed, so return nil.
      Utility.logger.info('no signatures removed; returning nil')
      return nil
    else
      # At least one signature node removed.
      # Note that we print from root to avoid the <?xml ...> junk.
      Utility.logger.info("removed #{counter} signatures removed; returning stripped metadata")
      return xml_doc.root.to_s
    end
  end

  # Return the metadata but with the <?xml ...> removed.
  def self.strip_xml_declaration(metadata)
    xml_doc = Nokogiri::XML(metadata) do |config|
      config.strict.noblanks
    end
    return xml_doc.root.to_s
  end


  # Because REMOTE_USER is only accessible from the request object, it is
  # set as a GLOBAL variable ($remote_user) in the application controller.
  # rubocop:disable Style/GlobalVars
  def self.get_REMOTE_USER()
    if ($remote_user.nil?() or $remote_user.empty?()) then
      raise 'REMOTE_USER is not set?!?'
    else
      return $remote_user
    end
  end
  # rubocop:enable Style/GlobalVars

  # Return the REMOTE_USER environment variable if set, otherwise return the
  # owner_auto_administrator value.
  def self.get_REMOTE_USER_or_autoadmin()
    remote_user = Utility.get_REMOTE_USER()

    if (remote_user.nil?() or remote_user.empty?()) then
      remote_user = Rails.configuration.owner_auto_administrator
    end

    return remote_user
  end

  # Return true if <sunetid> is an SPDB admin, false otherwise
  def self.admin?(sunetid)
    # We want 'auto-migration' to act as an admin.
    if (sunetid == 'auto-migration') then
      return true
    else
      return Rails.configuration.app_admin_sunetids.include?(sunetid)
    end
  end

  # If current user is an SPDB admin or is in one of the
  # groups that is allowed to write, return true. If one of the
  # groups allowed to write is "*" this means _everyone_ gets to write.
  def self.can_write?()
    remote_user = Utility.get_REMOTE_USER()
    if (Utility.admin?(remote_user)) then
      return true
    end

    epes = Utility.get_eduPersonEntitlements()

    allowed_groups = Rails.configuration.app_groups_that_can_write

    allowed_groups.each do |allowed_group|
      if (allowed_group == "*") then
        return true
      elsif (epes.include?(allowed_group)) then
        return true
      end
    end

    # If we get here then none of the user's eppns is allowed. That
    # means this user is _not_ allowed to write.
    return false
  end

  # Returns an array of eduPersonEntitlement's. For example:
  #  ['acs:team', 'stanford:stanford']
  def self.get_eduPersonEntitlements()
    if ($remote_env.key?('eduPersonEntitlement')) then
      epes_string = $remote_env['eduPersonEntitlement']
      epes = epes_string.split(/;/)
    else
      epes = []
    end

    return epes
  end

  # Return true if <sunetid> is an SPDB "read-only" admin, false otherwise
  def self.readonly_admin?(sunetid)
    return Rails.configuration.app_readonly_admin_sunetids.include?(sunetid)
  end


  # Sign an XML file. The key-pair used is set in the application.rb
  # configuration file.
  #
  # This command OVERWRITES the file specified by the parameter <signed_file>.
  #
  # Returns: stdout, stderr, and status of signing command.
  def self.sign_xml_file(unsigned_file, signed_file,
                         signing_key = Rails.configuration.sp_metadatafile_sign_private,
                         signing_certificate = Rails.configuration.sp_metadatafile_sign_public)
    ENV['JAVA_HOME'] = '/usr/lib/jvm/default-java'

    # Write the signed file to a temporary locatation and then move it
    # into place.

    tmp_signed_file      = Tempfile.new('xmlsigning')
    tmp_signed_file_path = tmp_signed_file.path()
    tmp_signed_file.close()

    cmds = ['/usr/share/xmlsectool/xmlsectool.sh',
            '--sign',
            '--inFile',      unsigned_file,
            '--outFile',     tmp_signed_file_path,
            '--key',         signing_key,
            '--certificate', signing_certificate,
            '--referenceIdAttributeName', 'ID',
           ]

    cmd = cmds.join(' ')

    stdout, stderr, status = Open3.capture3(cmd)

    if (status.exitstatus != 0) then
      Utility.logger.error("signing file error: [stdout: #{stdout}][stderr: #{stderr}][status: #{status}]")
      return false
    else
      Utility.logger.info("signed metadata file and stored in '#{tmp_signed_file_path}'")
    end

    # Move signed XML file into place.
    FileUtils.mv(tmp_signed_file_path, signed_file)

    # Clean up
    if (File.exist?(tmp_signed_file_path)) then
      tmp_signed_file.unlink()
    end

    return true
  end

  # cmd should be a string.
  def self.run_cmd_under_service_principal(cmd)
    k5start_cmd = '/usr/bin/k5start'

    cmds = [
      k5start_cmd,
      '-U',
      '-f', '/etc/spdb/spdb.keytab',
      '--',
      'sh', '-c',
      cmd,
    ]

    cmd_joined = cmds.join(' ')
    Utility.logger.info("cmd is [#{cmd_joined}]")

    stdout, stderr, p_status = Open3.capture3(cmd_joined)
    rc = p_status.exitstatus()
    Utility.logger.debug("[#{stdout}][#{stderr}][#{p_status}][#{rc}]")

    return [stdout, stderr, rc]
  end




  def self.get_mailman_subscriber_list(list_name)
    cmd = %Q('remctl mailman mm who #{list_name}')
    (stdout, stderr, rc) = self.run_cmd_under_service_principal(cmd)

    if (rc == 0) then
      # Parse the output
      lines = stderr.split("\n")
      addresses = []
      lines.each do |line|
        # Each line is an address until the first blank line
        if (line =~ /\S/) then
          addresses.push(line.strip())
        else
          # If we get here there are no more.
          break
        end
      end
      Utility.logger.info("found #{addresses.length} addresses in the mailman list #{list_name}")
      Utility.logger.info("the addresses are #{addresses.join('q')}")
      return addresses
    else
      # Some other error. What should we do?
      Utility.logger.error("error running command #{cmd}: [#{stdout},#{stderr}]")
      return []
    end

  end

  def self.remove_email_from_partners_list(email_address)
    list_name = Rails.configuration.partners_list_name

    cmd = %Q('remctl mailman mm unsub #{list_name} #{email_address}')
    (stdout, stderr, rc) = self.run_cmd_under_service_principal(cmd)

    if (rc == 0) then
      Utility.logger.debug("successfully removed #{email_address} from #{list_name}")
      return true
    else
      # Some other error. What should we do?
      Utility.logger.error("error running command #{cmd}: [#{stdout},#{stderr}]")
      return false
    end
  end

  def self.remove_unused_email_from_partners_list(email_address)
    if (!email_address) then
      Utility.logger.error('missing address for remove_email_from_partners_list')
      return
    else
      # Find out if this address is used anywhere. If so, don't delete.
      number_with_address = Spconfig.where(contact_email: email_address).count
      Utility.logger.info("found #{number_with_address} spconfigs with contact_email equal to #{email_address}")
      if (number_with_address.positive?) then
        Utility.logger.info("found at least one spconfig with contact_email equal to #{email_address}")
        # Don't delete from partners list
      else
        Utility.logger.info("found no spconfigs with contact_email equal to #{email_address}")
        # Do delete from partners lists
        Utility.remove_email_from_partners_list(email_address)
      end
    end
  end


  # old = old contact_email address
  # new = old contact_email address
  # 1. If old is undefined, add new to mailman list.
  # 2. If old and new are equal do nothing.
  # 3. If old and new are not equal:
  #    3a. Add new.
  #    3b. Remove old (unless it is a contact email for another existing
  #        spconfig).

  def self.add_email_to_partners_mailman_list(email_address_new, email_address_old)
    if (!Rails.configuration.partners_list_add_to_list) then
      # We don't want to add to the partners list, so simply return.
      Utility.logger.info('skipping adding e-mail to partners list by configuration')
      return false
    end

    if (!email_address_new) then
      Utility.logger.warn('no e-mail address, so cannot add to partners list')
      return false
    end

    if (email_address_new == email_address_old) then
      Utility.logger.info('skipping adding e-mail (no change in address)')
      return false
    end

    list_name = Rails.configuration.partners_list_name

    cmd = %Q('remctl mailman mm sub #{list_name} #{email_address_new}')
    (stdout, stderr, rc) = self.run_cmd_under_service_principal(cmd)

    if (rc == 0) then
      Utility.logger.debug("successfully added #{email_address_new} to #{list_name}")
    else
      # Some other error. What should we do?
      Utility.logger.error("error running command #{cmd}: [#{stdout},#{stderr}]")
      return false
    end

    # Remove the old address if not used by any other spconfig..
    Utility.remove_unused_email_from_partners_list(email_address_old)

    return
  end

  # Download the InCommon metadata and put it in /var/tmp
  def self.get_incommon_metadata()
    url         = Rails.configuration.incommon_metadata_url
    destination = '/var/tmp/InCommon-metadata.xml'

    require 'open-uri'
    # Note that we open in binary mode since the XML contains
    # some UTF characters.
    File.open(destination, 'wb') do |file|
      file.write open(url).read
      end

    # Extract entityIDs from the file
    Utility.extract_entity_ids_from_file(destination)
  end

  def self.extract_entity_ids_from_file(filename)
    entityIDs = []
    File.readlines(filename).each do |line|
      if (m = line.match(/entityID="([^"]*)"/)) then
        entityID = m[1]
        entityIDs.push(entityID)
      end
    end

    # Write these to a file.
    destination = '/var/tmp/incommon-entity-ids.txt'
    File.open(destination, 'w') do |file|
      entityIDs.each do |entityID|
        file.puts entityID
      end
    end
  end

  # Returns true if is in the InCommon entityIDs file, false if not, and
  # raises an error on problems. If the incommon entity file does not
  # exists, return false. We do this in case that file gets deleted
  # accidentally we don't want to break the SPDB application.
  def self.is_an_incommon_entity_id_file(entityID)
    if (entityID.strip() == '') then
      raise 'entityID passed is blank'
    end

    incommon_file = '/var/tmp/incommon-entity-ids.txt'

    if (not File.exist?(incommon_file)) then
      Utility.logger.error("InCommmon entityIDs file '#{incommon_file}' does not exist")
      return false
    end

    cmds = ['grep',
            entityID,
            '/var/tmp/incommon-entity-ids.txt',
           ]
    cmd = cmds.join(' ')
    Utility.logger.debug("cmd is #{cmd}")

    stdout, stderr, status = Open3.capture3(cmd)
    stdout.strip!()

    # Get all the lines in case there is more than one match.
    Utility.logger.debug("stdout is [#{stdout}]")
    lines = stdout.split("\n")

    lines.each do |line|
      if (line == entityID) then
        return true
      end
    end

    return false
  end

  # Returns true if is in the InCommon via mdq. If not, return false and
  # raises an error on problems.
  def self.is_an_incommon_entity_id(entityID)
    entityid_enc = ""
    if (entityID.strip() == '') then
      raise 'entityID passed is blank'
    else
      entityid_enc = ERB::Util.url_encode(entityID)
    end

    incommon_base="https://mdq.incommon.org/entities/"

    cmds = ['curl -s -I --fail ',
            incommon_base,
            entityid_enc
           ]
    cmd = cmds.join('')
    Utility.logger.debug("cmd is #{cmd}")
    stdout, stderr, status = Open3.capture3(cmd) 

    if (status.exitstatus != 0) then
      return false
    else
      return true
    end

  end


  # Get the version string from the file config/version.
  def self.get_version()
    version_file = Rails.root.join('config', 'version').to_s()
    if (File.exist?(version_file)) then
      File.open(version_file, 'r') do |f|
        return f.readline
      end
    else
      return '(unknown)'
    end
  end

  # Submit a metadata string to the validation service and return the results.
  def self.submit_metadata_for_validation(metadata)
    # POST the metadata to the validation web service and get the
    # results.
    uri_string = Rails.configuration.saml_sp_validate_url
    uri = URI.parse(uri_string)

    http = Net::HTTP.new(uri.host, uri.port)

    if (uri_string =~ /^https/) then
      http.use_ssl = true
    else
      http.use_ssl = false
    end

    req = Net::HTTP::Post.new(uri.path)

    data_to_post = {
      'metadata' => metadata,
    }

    req.set_form_data(data_to_post)
    response = http.request(req)
    results  = response.body()
    metadata_validation_results = JSON.parse(results)

    # @sp_info                = @metadata_validation_results[0]
    # @saml_validation_errors = @metadata_validation_results[1]

    return metadata_validation_results
  end

  def self.get_ACS_elements(metadata)
    xml_doc = Nokogiri::XML(metadata).remove_namespaces!

    acs_hash = {}

    acs_nodes = xml_doc.xpath('//AssertionConsumerService')

    acs_nodes.each do |acs_node|
      index    = acs_node.xpath('@index').to_s
      location = acs_node.xpath('@Location').to_s
      binding  = acs_node.xpath('@Binding').to_s

      if (!index.nil?() && location && binding) then
        acs_hash[index] = [binding, location]
      end
    end

    return acs_hash
  end

  def self.get_Certificates(metadata)
    xml_doc = Nokogiri::XML(metadata).remove_namespaces!

    certs = []

    kd_nodes = xml_doc.xpath('/EntityDescriptor//KeyDescriptor')

    kd_nodes.each do |kd_node|
      use = kd_node.xpath('@use').to_s
      if (use.nil?() or use.empty?()) then
        use = 'unspecified'
      end

      # Get the certs
      cert_nodes = kd_node.xpath('./KeyInfo/X509Data/X509Certificate')

      cert_nodes.each do |cert_node|
        cert = cert_node.text()
        certs.push([use, cert])
      end
    end

    return certs
  end

  def self.get_certificate_objects_from_metadata(metadata)
    # Get the certificates (if any)
    certs = Utility.get_Certificates(metadata)

    # Convert to Certificate objects.
    errors = []
    certificates = []
    certs.each do |cert|
      use = cert[0]
      pem = cert[1]

      begin
        new_cert = Utility.cert_string_to_certificate(pem)
      rescue OpenSSL::X509::CertificateError => e
        raise 'Unable to parse one or more of the certificates in submitted metadata.'
      end

      new_cert.use = use
      certificates.push(new_cert)
    end

    return certificates
  end

  # Returns an array of strings with certificate information.
  def self.get_certificate_info(metadata)
    cert_information = []

    uri_string = Rails.configuration.saml_sp_validate_url
    uri        = URI.parse(uri_string)
    http       = Net::HTTP.new(uri.host, uri.port)
    if (uri_string =~ /^https/) then
      http.use_ssl = true
    else
      http.use_ssl = false
    end

    req = Net::HTTP::Post.new(uri.path)

    data_to_post = {
      'metadata' => metadata
    }

    req.set_form_data(data_to_post)
    response = http.request(req)
    results  = response.body()
    meta     = JSON.parse(results)
    curtime  = DateTime.now.strftime('%s').to_i

    # Get the KeyDescriptors element
    if (meta[0].key?('KeyDescriptors')) then
      key_descriptors = meta[0]['KeyDescriptors']

      # Iterate through all the KeyDescriptors
      key_descriptors.each do |key_descriptor|
        usage = nil
        if (key_descriptor.key?('usage')) then
          usage = key_descriptor['usage']
        else
          usage = 'not specified'
        end

        if (not usage or (usage == 'nil')) then
          usage = 'not specified'
        end

        # Look at all the certificates
        if (key_descriptor.key?('X509Certificates')) then
          x509Certificates = key_descriptor['X509Certificates']

          x509Certificates.each do |x509Certificate|
            cert_info = {}

            selfSigned         = x509Certificate['selfSigned']
            subject            = x509Certificate['subject']
            fingerprint_md5    = x509Certificate['fingerprint-md5']
            fingerprint_sha256 = x509Certificate['fingerprint-sha256']

            ## 1. The CN
            # Get the last part of the subject (i.e., the "CN=" part.
            chunks = subject.split(',')
            last_chunk = chunks[-1]

            # Remove the "XX=" part (e.g., "CN=")
            last_chunk.gsub!(/^[^=]+=/, '')

            cert_info['cn'] = last_chunk

            ## 2. The certificate expiration
            cert_info['expires'] = Time.at(x509Certificate['notAfter_epoch']).strftime('%m/%d/%Y')

            ## 3. The usage.
            cert_info['usage'] = usage

            ## 3. The fingerprints
            cert_info['fingerprint-md5']    = fingerprint_md5
            cert_info['fingerprint-sha256'] = fingerprint_sha256

            ## 4. Expiration warning.
            expiration_warning = nil
            if (Time.at(x509Certificate['notAfter_epoch']).to_i > curtime)
              if (curtime > Time.at(x509Certificate['notAfter_epoch']).to_i - 63072000)
                expiration_warning = 'Warning: certificate expires in less than 2 years'
                expiration_warning_level = 'warn'
              else
                expiration_warning = nil
              end
            else
              expiration_warning = 'Warning: this certificate has EXPIRED'
              expiration_warning_level = 'critical'
            end

            if (not expiration_warning.nil?) then
              cert_info['expiration_warning']       = expiration_warning
              cert_info['expiration_warning_level'] = expiration_warning_level
            end

            ## 5. Self-signed warning
            if (selfSigned == 0) then
              self_signed_warning = 'Warning: certificate is not self-signed'
              cert_info['self_signed_warning'] = self_signed_warning
            end

            cert_information.push(cert_info)
          end
        end
      end

    end

    return cert_information

  end

  # Remove any leading/trailing whitespace and replace line-endings with "\n"
  def self.normalize_whitespace(string)
    # 1. Remove any leading or trailing whitespace
    string1 = string.strip()

    # 2. Fix line endings
    string2 = string1.encode(string1.encoding, universal_newline: true)

    return string2
  end

  def self.first_level_breadcrumb
    return [ [Rails.application.config.first_breadcrumb, '/home'] ]
  end

  # Given a certificate in PEM format return a Certificate object. Note
  # that the "use" attribute is set to "unspecified".
  def self.cert_string_to_certificate(pem)
    pem = pem.strip()

    # Step 1. Make sure that the pem string has the
    # correct header and footer.
    if (pem !~ /\A-----BEGIN CERTIFICATE-----/) then
      pem = "-----BEGIN CERTIFICATE-----\n" + pem
    end
    if (pem !~ /-----END CERTIFICATE-----\z/) then
      pem += "\n-----END CERTIFICATE-----"
    end

    # Step 2. Parse the certificate using the built-in openssl library.
    cert_obj = OpenSSL::X509::Certificate.new(pem)

    # Step 3. Create a new Certificate object and populate its attributes.
    cert = Certificate.new()

    cert.pem                 = pem
    cert.use                 = 'unspecified'
    cert.issuer              = cert_obj.issuer.to_s()
    cert.serial_number       = cert_obj.serial.to_s()
    cert.version             = cert_obj.version
    cert.subject             = cert_obj.subject.to_s()
    cert.valid_at            = cert_obj.not_before
    cert.expires_at          = cert_obj.not_after
    cert.signature_algorithm = cert_obj.signature_algorithm

    der                     = cert_obj.to_der
    cert.fingerprint_md5    = Digest::MD5.hexdigest(der)
    cert.fingerprint_sha256 = Digest::SHA256.hexdigest(der)

    extensions = cert_obj.extensions
    extensions.each do |extension|
      ext_s = extension.to_s()
      if (ext_s =~ /authorityKeyIdentifier = keyid:([0-9a-f\:]+)[^0-9a-f\:]/i) then
        cert.authority_key_identifier = $1
      end
      if (ext_s =~ /subjectKeyIdentifier = (.*)$/i) then
        subject_key_identifier = $1
        # We skip the subject_key_identifier if it is longer than 60
        # characters (the maximum length of the database field) or
        # contains any non-printable characters
        if (subject_key_identifier.size > 60) then
          subject_key_identifier = nil
        elsif (not subject_key_identifier.ascii_only?) then
          subject_key_identifier = nil
        end
        cert.subject_key_identifier = subject_key_identifier
      end
    end

    # Step 4. In this section we determine what kind of key-pair: RSA,
    # DSA, or EC.
    public_key = cert_obj.public_key
    key_type = nil

    if (key_type.nil?())
      begin
        pkey = OpenSSL::PKey::RSA.new(public_key)
        key_type = 'RSA'
      rescue
        # Not RSA
      end
    end

    if (key_type.nil?())
      begin
        pkey = OpenSSL::PKey::DSA.new(public_key)
        key_type = 'DSA'
      rescue
        # Not DSA
      end
    end

    if (key_type.nil?())
      begin
        pkey = OpenSSL::PKey::EC.new(public_key)
        key_type = 'EC'
      rescue
        # Not EC
      end
    end

    if (!key_type.nil?())
      cert.key_algorithm = key_type
    end

    # Step 5. Set the "is_selfsiged" attribute.
    if (cert.issuer == cert.subject) then
      cert.is_selfsigned = true
    else
      cert.is_selfsigned = false
    end

    return cert
  end

  # Extract the CN from a subject
  def self.subject_to_cn(subject)
    if (subject =~ /CN=(.*?)([,\/]|$)/i) then
      return $1
    else
      # If we get here, we give up and just return the entire subject as CN.
      return subject
    end
  end


  # Return true if the specified file is due to be updated, false
  # otherwise.
  #
  # Here is the logic.
  #  IF force_update is true
  #    THEN the target file will be created/updated
  #  ELSE
  #    IF the target file DOES NOT exist
  #      THEN target file will be created.
  #    ELSE
  #      IF the last updated time in the updatetime table is NEWER than
  #      the target file
  #        THEN the target file will be updated
  #      ELSE
  #        IF the target file is older than time_limit_seconds
  #          THEN the target file will be updated
  #        ELSE
  #          The target file will NOT be updated.
  #
  def self.should_file_be_exported(filepath:,
                                   updatedtime_name:,
                                   force_update:       false,
                                   time_limit_seconds: 2 * 60 * 60,
                                   action_label:       'unspecified'
                                  )
    write_file = nil

    # ###     # ###     # ###     # ###     # ###     # ###     # ###     # ###
    logger_proc = Proc.new {|msg|
      Utility.logger.info "[should_file_be_exported]/#{action_label}: #{msg}"
    }
    # ###     # ###     # ###     # ###     # ###     # ###     # ###     # ###

    if (force_update) then
      # force_update is true, then write file.
      write_file = true
      logger_proc.call('force_update is true, so will create/update')
    else
      if (not File.exist?(filepath)) then
        # No existing file? Then write file.
        write_file = true
        logger_proc.call('file does not exist, so will create')
      else
        # The file exists. Is it older than the time in the updatedtime
        # table?
        last_updated_time = Updatedtime.last_updated(updatedtime_name)
        if (last_updated_time.nil?() or (File.mtime(filepath) < last_updated_time)) then
          # The existing file is older, so update.
          write_file = true
          logger_proc.call('file is older than last-update time, so will update')
        else
          # The file exists but is not older than the last updated time. How old is it?
          file_age_seconds = (Time.now - File.mtime(filepath)).to_i()

          # If older than time_limit_seconds, write file
          if (file_age_seconds > time_limit_seconds) then
            write_file = true
            logger_proc.call('file exists but is old enough to require an update')
          else
            logger_proc.call('file exists and does not need to be updated')
          end
        end
      end
    end

    return write_file
  end

  def self.run_command_ssh_agent(cmd, ssh_private_key_path)
    if (ssh_private_key_path.nil?() or ssh_private_key_path.empty?()) then
      raise 'missing ssh_private_key_path'
    end

    return Open3.capture3(
             'ssh-agent',
             'bash', '-c',
             "ssh-add #{ssh_private_key_path}; #{cmd}",
           )
  end

  def self.git_commit_and_push(git_dir, message)
    # Step 1. Add everything in case there is a new file.
    cmd = "git -C #{git_dir} add ."
    stdout, stderr, status = Open3.capture3(cmd)

    if (not stderr.empty?()) then
      raise stderr
    end


    # Step 2. Does anything need to be committed?
    cmd = "git -C #{git_dir} diff --exit-code && git -C #{git_dir} diff --cached --exit-code "
    stdout, stderr, status = Open3.capture3(cmd)

    need_to_commit = nil
    if (status.exitstatus.zero?) then
      # Nothing needs to be commited, so we can exit now.
      Utility.logger.info("[git_commit] #{git_dir}: nothing to commit")
      need_to_commit = false
    else
      need_to_commit = true
      Utility.logger.info("[git_commit] #{git_dir}: something needs to be committed")
    end

    if (need_to_commit) then
      # Step 3. Make the commit.
      ENV['GIT_AUTHOR_NAME']  = 'auto-commit'
      ENV['GIT_AUTHOR_EMAIL'] = 'root@localhost'
      if (message.nil?() or message.empty?())
        message = 'auto-commit'
      end

      cmd = "git -C #{git_dir} commit -m '#{message}'"
      stdout, stderr, status = Open3.capture3(cmd)
      # raise "status is #{status.exitstatus}"

      if (status.exitstatus != 0) then
        Utility.logger.error("[git_commit] #{git_dir}: error committing: #{stderr}")
        return
      else
        Utility.logger.info("[git_commit] #{git_dir}: make commit with message '#{message}'")
      end
    else
      Utility.logger.info('[git_commit] skipping commit as nothing to commit')
    end

    # We pull and push anyway, just in case.

    # The next two commands require ssh keys
    ssh_private_key_path = Rails.configuration.ssh_private_key_path

    # Step 4. Pull, but favor our version.
    cmd = "git -C #{git_dir} pull -X ours origin master"
    stdout, stderr, status = Utility.run_command_ssh_agent(cmd, ssh_private_key_path)
    if (status.exitstatus != 0) then
      Utility.logger.error("[git_commit] #{git_dir}: error pulling: #{stderr}")
      return
    else
      Utility.logger.info('[git_commit] pulled succesfully')
    end

    # Step 5. Push.
    cmd = "git -C #{git_dir} push origin master"
    stdout, stderr, status = Utility.run_command_ssh_agent(cmd, ssh_private_key_path)

    if (status.exitstatus > 1) then
      Utility.logger.error("[git_commit] #{git_dir}: error pushing: #{stderr}")
      return
    else
      Utility.logger.info("[git_commit] #{git_dir}: pushed successfully")
    end

    return
  end

  # This function takes as input the user's name and returns an array
  # of applications they are allowed to view. This function uses the
  # array config.application_visibility defined in application.rb to
  # figure this out.
  #
  # The application_visibility configuration setting affects both
  # what applications a user can see and what the HOME page looks
  # like.
  #
  # The application_visibility variable is a hash mapping an
  # application name to a visibility value. Each visiblity value
  # must be one of these:
  #
  #    'everyone': all users can see this application
  #
  #    'admin':    only the admin users in the array app_admin_sunetids
  #                can see this application
  #
  #    'disabled': this application is disabled and will not show up
  #                as a choice in the home page
  #
  #    an array: only the users in the array can see this application;
  #              note that SPDB admins will also have access even if they
  #              are not part of the array.
  #
  # Example 1.
  #
  #   config.application_visibility = {
  #     'spconfigs'   => 'everyone',
  #     'federations' => 'admin',
  #     'awsaccounts' => ['adamhl', 'vivienwu'],
  #     'idps'        => 'disabled',
  #   }
  #
  # This means:
  #  - everyone will be able to access the spconfigs application
  #  - only admins can access the federations application
  #  - the two users 'adamhl' and 'vivienwu' and any SPDB administrator
  #    can access the awsaccounts application
  #  - the idps application is not accessible by anyone
  #
  # ---------------------------------------------------------------------------
  #
  # Here is the current list of applications:
  #
  #  * spconfig
  #  * awsaccount
  #  * federation
  #  * idp
  #
  # The global variable $visible_applications contains an array of
  # applications that the current user is allowed to access. If, for
  # some reason, $visible_applications is the empty array, an error
  # will be raised.
  #
  # If $visible_applications contains exactly one application, the
  # home page for the user will be that applications home page. If
  # the $visible_applications contains more than application, the
  # home page will be a set of buttons, pne for each application.

  def self.visible_applications(username)
    Utility.logger.debug("visible_applications username is '#{username}'")
    application_visibility = Rails.configuration.application_visibility

    visible_apps = []
    for application in application_visibility.keys
      visibility = application_visibility[application]

      if (visibility.is_a?(String))
        if (visibility == 'everyone')
          Utility.logger.debug("#{application}: everyone")
          visible_apps.push(application)
        elsif (visibility == 'admin')
          Utility.logger.debug("#{application}: admin")
          if (Utility.admin?(username))
            visible_apps.push(application)
          else
            # No access.
          end
        elsif (visibility == 'disabled')
          Utility.logger.debug("#{application}:: disabled")
          # Don't push as this application is disabled.
        else
          raise "(string) unrecognized visibility '#{visibility}' for application '#{application}'"
        end
      elsif (visibility.is_a?(Array)) then
        # Remember that if the user is an SPDB admin they automatically
        # get access even if they are not listed in the array.
        if (Utility.admin?(username))
          Utility.logger.debug("#{application}:: allowed (user is an admin)")
          visible_apps.push(application)
        elsif (visibility.include?(username)) then
          Utility.logger.debug("#{application}:: allowed")
          visible_apps.push(application)
        else
          # They are not in the array, so they don't get pushed.
          Utility.logger.debug("#{application}:: not allowed")
        end
      else
        raise "(array) unrecognized visibility '#{visibility.inspect}' for application '#{application}'"
      end
    end

    return visible_apps
  end

  # Determine the home page for this user.
  #
  # Algorithm: Get a list of the applications this user has access to. If
  # the list is longer than 1 then this user gets to see "helps#home",
  # i.e., the page that shows the application buttons. Otherwise, they get
  # directed to their application's index page.
  def self.home_page(username)
    visible_applications = Utility.visible_applications(username)

    if (visible_applications.length() == 0) then
      raise 'this user has no visible applications; this should never happen'
    elsif (visible_applications.length() == 1) then
      application = visible_applications[0]
      route = application + '#index'
    else
      route = 'helps#home'
    end

    Utility.logger.info("route is '#{route}'")
    return route
  end

#  # Return the JSON string for the attributes for entityid and user.
#  def self.get_attributes_for_old(entity_id, uid)
#    # Get the resolver test URL
#    base_url = Rails.configuration.x.resolver_test.url
#
#    # URL-encode the entityid
#    entityid_url_encoded = URI.escape(entity_id)
#
#    # Get username and password from the
#    credentials_file = Rails.configuration.x.resolver_test.credentials_file
#    credentials = YAML.load(File.read(credentials_file))
#    username = credentials['username']
#    password = credentials['password']
#
#    # Construct URL
#    uri = URI("#{base_url}?requester=#{entityid_url_encoded}&principal=#{uid}")
#
#    # Make request
#    Net::HTTP.start(uri.host,
#                    uri.port,
#                    :use_ssl     => uri.scheme == 'https',
#                    :verify_mode => OpenSSL::SSL::VERIFY_NONE) do |http|
#
#      request = Net::HTTP::Get.new uri.request_uri
#      request.basic_auth username, password
#
#      response = http.request request # Net::HTTPResponse object
#
#      my_hash = JSON.parse(response.body)
#
#      return my_hash['attributes']
#    end
#  end

  def self.my_domain_name()
    case Rails.env
    when /prod/i
      spdb_domain = 'spdb.stanford.edu'
    when /dev/i
      spdb_domain = 'spdb-dev.stanford.edu'
    else
      spdb_domain = "spdb-#{Rails.env}.stanford.edu"
    end
    return spdb_domain
  end

  # Returns a hash with these keys:
  #   'Format':  NameID format fo the Subject element
  #   'Subject': The value of the Subject element
  #   'attributes': an array of the attribute values
  #
  # The 'attributes' array is an array of arrays. Each element
  # of the array has this format:
  #    [Name, FriendlyName, [values]]
  #
  #
  #  { 'Subject' => [nameid-format, value],
  #    'Attributes' =>
  #     {
  #      Name1 => [FriendlyName1, [values1]],
  #      Name2 => [FriendlyName2, [values2]],
  #        ...
  #     }
  #  }

  def self.get_attributes_for(entity_id, uid)
    # Get the resolver test URL
    base_url = Rails.configuration.x.resolver_test.url

    # URL-encode the entityid
    entityid_url_encoded = CGI.escape(entity_id)

    # Get username and password from the
    credentials_file = Rails.configuration.x.resolver_test.credentials_file
    credentials = YAML.load(File.read(credentials_file))
    username = credentials['username']
    password = credentials['password']

    # Construct URL
    uri = URI("#{base_url}?saml2=x&requester=#{entityid_url_encoded}&principal=#{uid}")

    # Make request
    Net::HTTP.start(uri.host,
                    uri.port,
                    use_ssl:     uri.scheme == 'https',
                    verify_mode: OpenSSL::SSL::VERIFY_NONE) do |http|

      request = Net::HTTP::Get.new uri.request_uri
      request.basic_auth username, password

      response = http.request request # Net::HTTPResponse object

      xml_doc = Nokogiri::XML(response.body).remove_namespaces!()

      info = {}

      info['RawXML'] = response.body.to_s

      # Get the Subject
      subject = xml_doc.xpath('//Subject') # A NodeSet
      nameid  = nil
      if (subject) then
        nameid_raw = subject.xpath('//NameID')
        if (nameid_raw) then
          nameid_first = nameid_raw.first
          if (nameid_first) then
            nameid  = nameid_first['Format'].strip()
          end
        end

        subject_first_raw = subject.first
        if (subject_first_raw) then
          subject_content = subject_first_raw.content.strip()
        else
          subject_content = nil
        end

        info['Subject'] = [nameid, subject_content]
      end

      # Get all the Attributes
      attributes = {}

      cur_attributes = xml_doc.search('Attribute') # A NodeSet
      cur_attributes.each do |attribute|
        # Get the name
        attribute_name          = attribute['Name']
        attribute_friendly_name = attribute['FriendlyName']

        if (not attribute_friendly_name) then
          attribute_friendly_name = nil
        end

        # Get the AttributeValues
        values = attribute.search('AttributeValue')
        attribute_values = []
        values.each do |value|
          value_string = value.content
          attribute_values.push(value.content)
        end

        if (attribute_friendly_name) then
          sort_name = attribute_friendly_name
        else
          sort_name = attribute_name
        end

        # We have all the attributes now, so push
        attributes[sort_name] = [attribute_name, attribute_friendly_name, attribute_values]
      end

      info['Attributes'] = attributes

      return info
    end
  end

  def self.make_help_request_url(entity_id, extra)
    entity_id_urlencoded = CGI.escape(entity_id)
    extra_urlencoded     = CGI.escape(extra)

    help_request_url = Rails.configuration.help_request_url.dup

    help_request_url.gsub!('%%ENTITYID%%', entity_id_urlencoded)
    help_request_url.gsub!('%%EXTRA%%',    extra_urlencoded)

    return help_request_url
  end

  # Use xmllint to verify that XML file is valid.
  def self.valid_xml?(path_to_xml_file)
    (stdout, stderr, rc) = Open3.capture3(
             'xmllint',
             '--noout',
             path_to_xml_file,
           )

    if (!stderr.empty?()) then
      # Nope. Bad XML.
      Utility.logger.error("xmllint stderr: '#{stderr}'")
      return false
    elsif (rc != 0)
      # Nope. Bad XML.
      Utility.logger.error("xmllint rc: '#{rc}'")
      return false
    else
      return true
    end

  end

  # Return true if the passed in entityid is valid. This means that it is an
  # http(s) URL or is one of the exceptions.
  def self.entityid_valid?(entityid)
    # If entityid is empty it is certainly not valid.
    if (entityid.empty?())
      return false
    end

    # Is this entityid in the exception list?
    # So far, the only exception we have is for AWS's cognito which
    # insistes on using a urn.
    if (entityid =~ /^(urn:amazon:cognito:sp).*$/) then
      return true
    end

    # If we get here, the entityid needs to be a URL.
    validURI = true

    begin
      uri = URI.parse(entityid)
    rescue URI::InvalidURIError
      validURI = false
    end

    # Do our own check.
    if (entityid !~ /^(http|https):\/\/.+/) then
      validURI = false
    end

    if (uri.host.nil?) then
      validURI = false
    end

    return validURI
  end

  def self.clear_cache()
    Rails.cache.clear()
  end

  def self.incommon_metadata_url(entityID)
    prefix = 'https://mdq.incommon.org/entities'
    entityID_encoded = CGI.escape(entityID)
    return "#{prefix}/#{entityID_encoded}"
  end

  # Remove duplicate objects which has same key 
  def self.delete_duplicate_objects(list_of_objects, key)
    objects_hash = {}

    list_of_objects.each do |obj|
    objname = obj.read_attribute(key)
      if (objects_hash.key?(objname)) then
        # This is already in the list, so delete
        obj.delete()
        Utility.logger.debug("deleting duplicate #{key} '#{objname}'")
      else
        objects_hash[objname] = true
      end
    end
  end 

  # Return string from an array, join by ','
  def self.get_objects_as_string(list_of_objects)
    the_string=list_of_objects.join(', ')
    if (the_string.blank?()) then
      the_string="none" 
    end 
    return "#{the_string}"
  end 

  # Return uniq workgroup names from groups table
  def self.distinct_workgroups()
    uniq_groups = []
    uniq_groups = Group.distinct.pluck(:workgroup)

    return uniq_groups
  end

  def self.get_all_my_privgroups(username)
    ldap_server     = Rails.configuration.ldap_service_name
    username_DN         = "-b cn=people,dc=stanford,dc=edu "
    filter_cmd      = "uid=#{username}"  
    filter_attr     = 'suPrivilegeGroup'
    ldapsearch_cmd  = '/usr/bin/ldapsearch'
    post_process_cmd =' | /bin/egrep ^suPrivilegeGroup: | /usr/bin/sort -u '
    my_privgroups   = []
    # The command looks like this:
    #
    #   /usr/bin/ldapsearch -Q -LLL -h ldap.stanford.edu \
    #   cn=people,dc=stanford,dc=edu uid=adamhl suPrivilegeGroup |egrep ^suPrivilegeGroup: |sort -u|cut -d' ' -f2 
    #
    #
    cmd = %Q('#{ldapsearch_cmd} -Q -LLL -h #{ldap_server} #{username_DN} "#{filter_cmd}" #{filter_attr} #{post_process_cmd} ' )
    #Utility.logger.debug("YYY ldapsearch cmd: #{cmd}")

    (stdout, stderr, rc) = Utility.run_cmd_under_service_principal(cmd)
    if (rc == 0) then
      lines = stdout.split("\n")
      lines.each do |line|
        if not (line.include?('/administrator'))
          group_name_array = line.split(' ').slice(1,1).map(&:strip) 
          group_name =  Utility.get_objects_as_string(group_name_array)
          my_privgroups.push(group_name)
        end 
      end #### eof for do loop
      return my_privgroups
    else
      Utility.logger.debug("rc not 0 ")
      raise 'Something has gone wrong... ldapsearch is not successful. Please use search function to find your sites or contact SPDB administrator.'
    end
  end
 
  #### return the intersect of all "distinct workgroups" from db  and user "all workgroup"
  def self.get_my_owning_workgroups(username)
    workgroups_in_db = []
    my_privgroups = []
    my_owning_workgroups = []

    my_privgroups = Utility.get_all_my_privgroups(username)
    workgroups_in_db = Utility.distinct_workgroups() 
    if ( !my_privgroups.empty? ) then
    ## intersect the above two arrays
      my_owning_workgroups = workgroups_in_db & my_privgroups 
      ##Utility.logger.debug("YYY my_owning_workgroups: #{my_owning_workgroups}")
    end
    return my_owning_workgroups 

  end
  # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###   # # ###
end
